"""Propagator classes that preserve the same shape."""

from jsonargparse import Namespace
from typing import List
from .base import BasePropagator, get_shape, create_shape


class SameShapePropagator(BasePropagator):
    """Propagator for blocks in which the input and output shapes are the same."""

    def initial_checks(self, from_blocks: List[Namespace], block: Namespace):
        """Method that does some initial checks before propagation.

        Calls the base class checks and if multi-input makes sure that all
        inputs have the same shape and if not multi-input makes sure that there
        is only a single input block.

        Args:
            from_blocks: The input blocks.
            block: The block to propagate its shapes.

        Raises:
            ValueError: When multi_input==False and len(from_blocks) != 1.
        """
        super().initial_checks(from_blocks, block)
        if self.num_input_blocks is not None:
            shape = get_shape('out', from_blocks[0])
            if not all(shape == get_shape('out', b) for b in from_blocks[1:]):
                in_shapes = ', '.join('[id='+b._id+']='+str(get_shape('out', b)) for b in from_blocks)
                raise ValueError(f'block[id={block._id}] of type {self.block_class} requires all inputs to '
                                 f'have the same shape, but got {in_shapes}.')
        else:
            if len(from_blocks) != 1:
                raise ValueError(f'block[id={block._id}] of type {self.block_class} only accepts one input '
                                 f'block, but got {len(from_blocks)}.')


    def propagate(self, from_blocks: List[Namespace], block: Namespace):
        """Method that propagates shapes to a block.

        Args:
            from_blocks: The input blocks.
            block: The block to propagate its shapes.
        """
        block._shape = create_shape(get_shape('out', from_blocks[0]))


class SameShapesPropagator(SameShapePropagator):
    """Propagator for blocks that receive multiple inputs of the same shape and preserves this shape."""

    num_input_blocks = '>1'


class SameShapeConsumeDimPropagator(SameShapePropagator):
    """Propagator for blocks in which the output shape is the same as input except the last which is consumed."""

    def initial_checks(self, from_blocks: List[Namespace], block: Namespace):
        """Method that does some initial checks before propagation.

        Calls the base class checks and makes sure that the input has more than
        one dimension.

        Args:
            from_blocks: The input blocks.
            block: The block to propagate its shapes.

        Raises:
            ValueError: When len(input_shape) < 2.
        """
        super().initial_checks(from_blocks, block)
        if len(get_shape('out', from_blocks[0])) < 2:
            raise ValueError(f'block[id={block._id}] of type {self.block_class} requires input to have '
                             f'at least two dimensions.')

    def propagate(self, from_blocks: List[Namespace], block: Namespace):
        """Method that propagates shapes to a block.

        Args:
            from_blocks: The input blocks.
            block: The block to propagate its shapes.
        """
        block._shape = create_shape(get_shape('out', from_blocks[0]), get_shape('out', from_blocks[0])[:-1])



"""Blocks definitions and functions related to registering propagators."""

import enum
from .propagators.concat import ConcatenatePropagator
from .propagators.conv import ConvPropagator, PoolPropagator
from .propagators.fixed import AddFixedPropagator, FixedOutputPropagator
from .propagators.group import SequentialPropagator, GroupPropagator
from .propagators.reshape import ReshapePropagator
from .propagators.rnn import RnnPropagator
from .propagators.same import SameShapePropagator, SameShapesPropagator, SameShapeConsumeDimPropagator
from .module import ModulePropagator


class SameShapeBlocksEnum(enum.Enum):
    """Enum of blocks that preserve the input shape."""

    Sigmoid = SameShapePropagator('Sigmoid')
    """Block that applies a sigmoid function."""

    LogSigmoid = SameShapePropagator('LogSigmoid')
    """Block that applies a log-sigmoid function."""

    Softmax = SameShapePropagator('Softmax')
    """Block that applies a softmax function."""

    LogSoftmax = SameShapePropagator('LogSoftmax')
    """Block that applies a log-softmax function."""

    Tanh = SameShapePropagator('Tanh')
    """Block that applies a hyperbolic tangent function."""

    ReLU = SameShapePropagator('ReLU')
    """Block that applies a rectified linear unit function."""

    LeakyReLU = SameShapePropagator('LeakyReLU')
    """Block that applies a leaky rectified linear unit function."""

    Dropout = SameShapePropagator('Dropout')
    """Block that applies dropout, randomly set elements to zero."""

    BatchNorm2d = SameShapePropagator('BatchNorm2d')
    """Block that does 2D batch normalization."""

    Identity = SameShapePropagator('Identity')
    """Block that does nothing, useful to connect one tensor to multiple blocks in a graph."""

    Add = SameShapesPropagator('Add')
    """Block that adds the values of all input tensors. Input tensors must have the same shape."""

    CRF = SameShapeConsumeDimPropagator('CRF')
    """A layer that performs CRF decoding."""


class ConcatBlocksEnum(enum.Enum):
    """Enum of blocks that concatenate multiple inputs."""

    Concatenate = ConcatenatePropagator('Concatenate')
    """Block that concatenates multiple inputs of the same shape along a given dimension."""


class FixedOutputBlocksEnum(enum.Enum):
    """Enum of blocks that have fixed outputs."""

    Linear = FixedOutputPropagator('Linear')
    """Linear transformation to the last dimension of input tensor."""

    Embedding = AddFixedPropagator('Embedding')
    """A lookup table that retrieves embeddings of a fixed size."""

    AdaptiveAvgPool1d = FixedOutputPropagator('AdaptiveAvgPool1d', unfixed_dims=1, fixed_dims=1)
    """1D adaptive average pooling over input."""

    AdaptiveAvgPool2d = FixedOutputPropagator('AdaptiveAvgPool2d', unfixed_dims=1, fixed_dims=2)
    """2D adaptive average pooling over input."""


class ConvBlocksEnum(enum.Enum):
    """Enum of convolution-style blocks."""

    Conv1d = ConvPropagator('Conv1d', conv_dims=1)
    """1D convolution."""

    Conv2d = ConvPropagator('Conv2d', conv_dims=2)
    """2D convolution."""

    Conv3d = ConvPropagator('Conv3d', conv_dims=3)
    """3D convolution."""

    MaxPool1d = PoolPropagator('MaxPool1d', conv_dims=1)
    """1D maximum pooling."""

    MaxPool2d = PoolPropagator('MaxPool2d', conv_dims=2)
    """2D maximum pooling."""

    MaxPool3d = PoolPropagator('MaxPool3d', conv_dims=3)
    """3D maximum pooling."""

    AvgPool1d = PoolPropagator('AvgPool1d', conv_dims=1)
    """1D average pooling."""

    AvgPool2d = PoolPropagator('AvgPool2d', conv_dims=2)
    """2D average pooling."""

    AvgPool3d = PoolPropagator('AvgPool3d', conv_dims=3)
    """3D average pooling."""


class RnnBlocksEnum(enum.Enum):
    """Enum of recurrent-style blocks."""

    RNN = RnnPropagator('RNN')
    """A simple recurrent block."""

    LSTM = RnnPropagator('LSTM')
    """An LSTM recurrent block."""

    GRU = RnnPropagator('GRU')
    """A GRU recurrent block."""


class ReshapeBlocksEnum(enum.Enum):
    """Enum of blocks that transform the shape."""

    Reshape = ReshapePropagator('Reshape')
    """Transformation of the shape of the input."""


class GroupPropagatorsEnum(enum.Enum):
    """Enum of blocks that group other blocks."""

    Sequential = SequentialPropagator('Sequential')
    """Sequence of blocks that are connected in the given order."""

    Group = GroupPropagator('Group')
    """Group of blocks with connected according to a given graph."""

    Module = ModulePropagator('Module')
    """Definition of a complete module."""


known_propagators = [
    SameShapeBlocksEnum,
    ConcatBlocksEnum,
    FixedOutputBlocksEnum,
    ConvBlocksEnum,
    RnnBlocksEnum,
    ReshapeBlocksEnum,
    GroupPropagatorsEnum,
]


propagators = {}


def register_propagator(propagator, replace: bool = False):
    """Adds a propagator to the dictionary of registered propagators."""
    if not replace and propagator.block_class in propagators:
        raise ValueError(f'Propagator for blocks of type {propagator.block_class} already registered.')
    invalid_classes = {'Default', 'Input', 'Output', 'Nested', 'Shared'}
    if propagator.block_class in invalid_classes or propagator.block_class.startswith('Nested'):
        raise ValueError(f'Propagators are not allowed to have as class any of {invalid_classes}.')
    propagators[propagator.block_class] = propagator


def register_known_propagators():
    """Function that registers all propagators defined in the modules of the package."""
    for propagators_enum in known_propagators:
        for propagator in propagators_enum:
            assert propagator.name == propagator.value.block_class
            register_propagator(propagator.value)


register_known_propagators()



"""Classes related to rendering of architectures."""

import os
import re
import itertools
import textwrap
from importlib.util import find_spec
from jsonargparse import ActionJsonSchema, ActionOperators, Namespace, namespace_to_dict, Path
from typing import Union
from .propagators.base import get_shape
from .propagators.group import get_blocks_dict, add_ids_prefix
from .module import ModuleArchitecture
from .graph import parse_graph
from .schemas import id_separator
from .sympy import sympify_variable


pygraphviz_available = find_spec('pygraphviz')


class ModuleArchitectureRenderer(ModuleArchitecture):
    """Class for instantiating a ModuleArchitectureRenderer objects useful for creating module architecture diagrams."""

    @staticmethod
    def get_config_parser():
        """Returns a ModuleArchitectureRenderer configuration parser."""
        parser = ModuleArchitecture.get_config_parser()
        parser.description = ModuleArchitectureRenderer.__doc__

        # render options #
        group_render = parser.add_argument_group('Rendering related options')
        group_render.add_argument('--save_pdf',
            default=False,
            type=bool,
            help='Whether to write rendered pdf file to output directory.')
        group_render.add_argument('--save_gv',
            default=False,
            type=bool,
            help='Whether to write graphviz file to output directory.')
        group_render.add_argument('--block_attrs',
            default={'Default':  'shape=box',
                     'Input':    'shape=box, style=rounded, penwidth=1.5',
                     'Output':   'shape=box, style=rounded, peripheries=2',
                     'Nested':   'shape=box, style=dashed',
                     'Shared':   'style=filled',
                     'Reshape':  'shape=hexagon',
                     'Identity': 'shape=circle, width=0',
                     'Add':      'shape=circle, margin=0, width=0'},
            action=ActionJsonSchema(schema={'type': 'object', 'items': {'type': 'string'}}),
            help='Attributes for block nodes.')
        group_render.add_argument('--block_labels',
            default={'Identity': '',
                     'Add':      '+'},
            action=ActionJsonSchema(schema={'type': 'object', 'items': {'type': 'string'}}),
            help='Fixed labels for block nodes.')
        group_render.add_argument('--edge_attrs',
            default='fontsize=10',
            help='Attributes for edges.')
        group_render.add_argument('--nested_depth',
            default=3,
            action=ActionOperators(expr=('>=', 0)),
            help='Maximum depth for nested subblocks to render. Set to 0 for unlimited.')
        group_render.add_argument('--full_ids',
            default=False,
            type=bool,
            help='Whether block IDs should include parent prefix.')
        group_render.add_argument('--layout_prog',
            choices=['dot', 'neato', 'twopi', 'circo', 'fdp'],
            default='dot',
            help='The graphviz layout method to use.')

        return parser


    def apply_config(self, cfg: Union[str, dict, Namespace]):
        """Applies a configuration to the ModuleArchitectureRenderer instance.

        Args:
            cfg: Path to config file or config object.
        """
        super().apply_config(cfg)

        if hasattr(cfg, 'block_attrs'):  # @todo support also dict
            block_attrs = {}
            for block, attrs in vars(cfg.block_attrs).items():
                attrs_dict = {}
                for a, v in [x.split('=') for x in re.split(', *', attrs)]:
                    attrs_dict[a] = v
                block_attrs[block] = attrs_dict
            if 'Default' not in block_attrs:
                block_attrs['Default'] = {'shape': 'box'}
            self.block_attrs = block_attrs

        if hasattr(cfg, 'edge_attrs'):
            edge_attrs = {}
            for a, v in [x.split('=') for x in re.split(', *', cfg.edge_attrs)]:
                edge_attrs[a] = v
            self.edge_attrs = edge_attrs


    @staticmethod
    def _set_architecture_description(graph, architecture):
        """Sets the architecture description to a graph as a label."""
        if hasattr(architecture, '_description'):
            description = architecture._description
            description = '<BR />'.join(textwrap.wrap(description, width=100))
            graph.graph_attr['label'] = f'<{description}>'
            graph.graph_attr['labelloc'] = 't'
            graph.graph_attr['labeljust'] = 'l'


    @staticmethod
    def _set_node_description(graph, node, full_ids=False):
        """Sets a node description as a label."""
        node_id = node._id if full_ids else node._id.split(id_separator)[-1]
        description = node_id
        if hasattr(node, '_description'):
            description = '<BR />'.join(textwrap.wrap(node._description, width=50))
            description = f'<{node_id}<FONT POINT-SIZE="6"><BR />{description}</FONT>>'
        graph.get_node(node._id).attr['label'] = description


    def _set_edge_label(self, graph, blocks, node_from, node_to, subblock=False):
        """Sets the shape dimensions to an edge as its label."""
        block_from = blocks[node_from]
        if hasattr(block_from, '_shape'):
            shape = get_shape('out', block_from)
            shape = ' × '.join(str(sympify_variable(d)) for d in shape)
            edge = graph.get_edge(node_from, node_to)
            edge.attr['label'] = ' '+shape
            edge_attrs = self.edge_attrs
            for a, v in edge_attrs.items():
                edge.attr[a] = v


    def _set_block_label(self, graph, block, graph_attr=False, full_ids=False):
        """Sets a block's label including its id and properties."""

        if hasattr(self.cfg.block_labels, block._class):
            label = getattr(self.cfg.block_labels, block._class)

        else:
            exclude = {'output_feats', 'graph', 'input', 'output', 'architecture'}
            name = block._class
            if hasattr(block, '_name'):
                name = block._name
            props = ''
            if hasattr(block, '_id'):
                block_id = block._id if full_ids else block._id.split(id_separator)[-1]
                props += f'<BR />id: {block_id}'
            for key in {'_path', '_id_share'}:
                if hasattr(block, key):
                    props += f'<BR />{key[1:]}: {getattr(block, key)}'

            def norm_prop(val):
                if isinstance(val, Namespace):
                    val = namespace_to_dict(val)
                elif isinstance(val, list):
                    val = [namespace_to_dict(v) if isinstance(v, Namespace) else v for v in val]
                return str(val)

            for k, v in vars(block).items():
                if not k.startswith('_') and k not in exclude:
                    if block._class in {'Sequential', 'Group'} and k == 'blocks':
                        props += f'<BR />{k}: {len(v)}'
                    else:
                        props += f'<BR />{k}: {norm_prop(v)}'
            if props != '':
                label = f'<{name}<FONT POINT-SIZE="6">{props}</FONT>>'
            else:
                label = name

        if graph_attr:
            graph.graph_attr['label'] = label
        else:
            graph.get_node(block._id).attr['label'] = label


    def _set_block_attrs(self, graph, blocks, block_class=None, graph_attr=False):
        """Sets graph style attributes to a block."""
        block_attrs = self.block_attrs
        for block in blocks:
            attrs_class = block._class if block_class is None else block_class
            attrs = block_attrs[attrs_class] if attrs_class in block_attrs else block_attrs['Default']
            for a, v in attrs.items():
                if graph_attr:
                    graph.graph_attr[a] = v
                else:
                    graph.get_node(block._id).attr[a] = v
            if hasattr(block, '_id_share') and 'Shared' in block_attrs:
                for a, v in block_attrs['Shared'].items():
                    if graph_attr:
                        graph.graph_attr[a] = v
                    else:
                        graph.get_node(block._id).attr[a] = v


    def create_graph(self):
        """Creates a pygraphviz graph of the architecture using the current configuration."""
        architecture = self.architecture
        blocks = self.blocks

        ## Create raw graph ##
        if not pygraphviz_available:
            raise ImportError('pygraphviz package is required by create_graph method.')
        from pygraphviz import AGraph
        graph = AGraph('\n'.join(['digraph {']+architecture.graph+['}']))

        ## Add architecture description ##
        self._set_architecture_description(graph, architecture)

        ## Set attributes of blocks ##
        self._set_block_attrs(graph, architecture.inputs, block_class='Input')
        self._set_block_attrs(graph, architecture.outputs, block_class='Output')
        self._set_block_attrs(graph, architecture.blocks)

        ## Add input/output descriptions ##
        for node in itertools.chain(architecture.inputs, architecture.outputs):
            self._set_node_description(graph, node)

        ## Add tensor shapes to edges ##
        for node_from, node_to in graph.edges():
            self._set_edge_label(graph, blocks, node_from, node_to)

        ## Set block properties ##
        for block in architecture.blocks:
            self._set_block_label(graph, block)

        ## Create subgraphs ##
        self._add_subgraphs(graph, architecture.blocks, dict(blocks), depth=2)

        return graph


    def _add_subgraphs(self, graph, blocks, subblocks_dict, depth, parent_graph=None):
        """Adds subgraphs to a graph if the depth is not higher that configured value."""
        if depth > self.cfg.nested_depth and not self.cfg.nested_depth == 0:
            return
        if parent_graph is None:
            parent_graph = graph
        full_ids = self.cfg.full_ids
        for block in [b for b in blocks if b._class in {'Sequential', 'Group', 'Module'}]:
            ## Remove edges and node ##
            edges = graph.edges(block._id)
            edges_from = [(u, v) for u, v in edges if v == block._id]
            edges_to = [(u, v) for u, v in edges if u == block._id]
            for edge in edges:
                graph.remove_edge(*edge)
            graph.remove_node(block._id)
            ## Create subgraph cluster ##
            subgraph = parent_graph.add_subgraph(name='cluster_'+block._id, labeljust='r', labelloc='t')
            self._set_block_label(subgraph, block, graph_attr=True, full_ids=full_ids)
            self._set_block_attrs(subgraph, [block], block_class='Nested', graph_attr=True)
            ## Handle Module ##
            if block._class == 'Module':
                subblocks_dict.update(get_blocks_dict(block.architecture.inputs+block.architecture.outputs))
                input_id = block.architecture.inputs[0]._id
                subgraph.add_node(input_id)
                self._set_node_description(graph, subblocks_dict[input_id], full_ids=full_ids)
                self._set_block_attrs(graph, [subblocks_dict[input_id]], block_class='Input')
                graph.add_edge(edges_from[0][0], input_id)
                self._set_edge_label(graph, subblocks_dict, edges_from[0][0], input_id, subblock=True)
                output_id = block.architecture.outputs[0]._id
                subgraph.add_node(output_id)
                self._set_node_description(graph, subblocks_dict[output_id], full_ids=full_ids)
                self._set_block_attrs(graph, [subblocks_dict[output_id]], block_class='Output')
                graph.add_edge(output_id, edges_to[0][1])
                self._set_edge_label(graph, subblocks_dict, output_id, edges_to[0][1], subblock=True)
                block = block.architecture
                edges_from[0] = (input_id, edges_from[0][1])
                edges_to = []
            ## Add subblocks nodes and edges ##
            if not self.cfg.propagated:
                add_ids_prefix(block, [])
            subblocks_dict.update(get_blocks_dict(block.blocks))
            blocks_from = [subblocks_dict[edges_from[0][0]]]
            topological_predecessors = parse_graph(blocks_from, block)
            for subblock_id, prev_ids in topological_predecessors.items():
                subgraph.add_node(subblock_id)
                subblock = subblocks_dict[subblock_id]
                if hasattr(subblock, '_class'):
                    self._set_block_label(subgraph, subblock, full_ids=full_ids)
                for node_id_prev in prev_ids:
                    graph.add_edge(node_id_prev, subblock_id)
                    self._set_edge_label(graph, subblocks_dict, node_id_prev, subblock_id, subblock=True)
            self._set_block_attrs(graph, block.blocks)
            ## Add final edges ##
            for u, v in edges_to:
                graph.add_edge(subblock_id, v)
                self._set_edge_label(graph, subblocks_dict, subblock_id, v, subblock=True)
            ## Add subgraphs ##
            self._add_subgraphs(graph, block.blocks, subblocks_dict, depth=depth+1, parent_graph=subgraph)


    def render(
        self,
        architecture: Union[str, Path] = None,
        out_render: Union[str, Path] = None,
        cfg: Namespace = None,
    ):
        """Renders the architecture diagram optionally writing to the given file path.

        Args:
            architecture: Path to a jsonnet architecture file.
            out_render: Path where to write the rendered diagram with a valid extension for pygraphviz
                        to determine the type.
            cfg: Configuration to apply before rendering.

        Returns:
            AGraph: pygraphviz graph object.
        """
        if cfg is not None:
            self.apply_config(cfg)
        if architecture is not None:
            self.load_architecture(architecture)
        graph = self.create_graph()
        outdir = self.cfg.outdir if isinstance(self.cfg.outdir, str) else self.cfg.outdir()
        if self.cfg.save_gv:
            out_gv = os.path.join(outdir, self.architecture._id + '.gv')
            self._check_overwrite(out_gv)
            graph.write(out_gv)
        graph.layout(prog=self.cfg.layout_prog)
        if self.cfg.save_pdf:
            out_pdf = os.path.join(outdir, self.architecture._id + '.pdf')
            self._check_overwrite(out_pdf)
            graph.draw(out_pdf)
        if out_render is not None:
            if not isinstance(out_render, str):
                out_render = out_render()
            self._check_overwrite(out_render)
            graph.draw(out_render)
        return graph



"""Functions related to the parsing of graphs."""

from collections import OrderedDict
from jsonargparse import dict_to_namespace, Namespace
from networkx import DiGraph
from networkx.algorithms.dag import is_directed_acyclic_graph, topological_sort
from typing import Dict, List


def digraph_from_graph_list(graph_list):
    edges = []
    for line in graph_list:
        nodes = line.split(' -> ')
        for num in range(1, len(nodes)):
            edges.append((nodes[num-1], nodes[num]))
    graph = DiGraph()
    graph.add_edges_from(edges)
    return graph


def parse_graph(from_blocks: List[Namespace], block: Namespace) -> Dict[str, List[str]]:
    """Parses a graph of a block.

    Args:
        from_blocks: The input blocks.
        block: The block to parse its graph.

    Returns:
        Dictionary in topological order mapping node IDs to its respective input nodes IDs.

    Raises:
        ValueError: If there are problems parsing the graph.
        ValueError: If the graph is not directed and acyclic.
        ValueError: If topological sort does not include all nodes.
    """
    if isinstance(block, dict):
        block = dict_to_namespace(block)
    if any(isinstance(x, dict) for x in from_blocks):
        from_blocks = [dict_to_namespace(x) for x in from_blocks]

    ## Get graph list ##
    if hasattr(block, '_class') and block._class == 'Sequential':
        graph_list = [from_blocks[0]._id + ' -> ' + ' -> '.join([b._id for b in block.blocks])]
    else:
        graph_list = block.graph
        if hasattr(block, 'input') and isinstance(block.input, str):
            graph_list = [from_blocks[0]._id+' -> '+block.input] + graph_list

    ## Parse graph ##
    try:
        graph = digraph_from_graph_list(graph_list)
    except Exception as ex:
        raise ValueError(f'Problems parsing graph for block[id={block._id}]: {ex}') from ex
    if not is_directed_acyclic_graph(graph):
        raise ValueError(f'Expected graph to be directed and acyclic for block[id={block._id}], graph={graph_list}.')

    ## Create topologically ordered dict mapping all nodes to its inputs ##
    topological_predecessors = OrderedDict()
    for node in topological_sort(graph):
        predecessors = [n for n in graph.predecessors(node)]
        if len(predecessors) > 0:
            topological_predecessors[node] = predecessors

    nodes_blocks = {b._id for b in block.blocks}
    nodes_topological = {k for k in topological_predecessors.keys()}
    missing = nodes_blocks - nodes_topological
    if len(missing) > 0:
        raise ValueError(f'Graph in block[id={block._id}] does not reference all of its blocks: missing={missing}.')

    return topological_predecessors



#!/usr/bin/env python3
"""Unit tests for instantiators."""

# pylint: disable=no-member

import os
import shutil
import tempfile
import unittest
from copy import deepcopy
from narchi.instantiators.common import import_object
from narchi.module import ModuleArchitecture
from narchi.schemas import auto_tag
from narchi_tests.data import *

try:
    import torch
    from torch.nn.utils.rnn import pad_packed_sequence
except:
    torch = False

if torch:
    from narchi.instantiators.pytorch import BaseModule, StandardModule, Reshape, standard_pytorch_blocks_mappings
    from narchi.instantiators.pytorch_packed import (PackedModule, packed_pytorch_blocks_mappings, pack_2d_sequences,
                                                     Conv2dPacked, MaxPool2dPacked)


@unittest.skipIf(not torch, 'torch package is required')
class PytorchTests(unittest.TestCase):
    """Tests for pytorch instantiator."""

    def test_base_module(self):
        mappings = dict(standard_pytorch_blocks_mappings)
        del mappings['Conv2d']

        class TestModule(BaseModule):
            blocks_mappings = mappings

        with torch.no_grad():
            self.assertRaises(NotImplementedError, lambda: TestModule(resnet_jsonnet, cfg=resnet_cfg))


    def test_reshape(self):
        ## successes ##
        examples = [
            {
                'in_shape': [1, 16, 8, 4, 2],
                'reshape_spec': [3, 2, 0, 1],
                'expected': [1, 2, 4, 16, 8],
            },
            {
                'in_shape': [2, 12, 8, 7, 2],
                'reshape_spec': [[3, 2], [0, 1]],
                'expected': [2, 14, 96],
            },
            {
                'in_shape': [7, 4608],
                'reshape_spec': [{'0': [3, 48, auto_tag]}],
                'expected': [7, 3, 48, 32],
            },
            {
                'in_shape': [1, 4608, 9],
                'reshape_spec': [1, {'0': [3, auto_tag, 32]}],
                'expected': [1, 9, 3, 48, 32],
            },
        ]
        for example in examples:
            reshape = Reshape(example['reshape_spec'])
            x = torch.rand(*example['in_shape'])
            y = reshape(x)
            self.assertEqual(list(y.shape), example['expected'])

        ## failures ##
        examples = [
            {
                'in_shape': [1, 4608, 9],
                'reshape_spec': [{'0': [3, 48, auto_tag]}],
            },
        ]
        for example in examples:
            reshape = Reshape(example['reshape_spec'])
            x = torch.rand(*example['in_shape'])
            self.assertRaises(RuntimeError, lambda: reshape(x))


    def test_resnet(self):
        cfg = deepcopy(resnet_cfg)

        with torch.no_grad():
            module = StandardModule(resnet_jsonnet, cfg=cfg)
            module.eval()
            image = torch.rand(1, 3, 256, 256)
            logits = module(image=image)
            self.assertEqual(list(logits.shape), [1, 1000])

            imagenet_jsonnet = os.path.join(data_dir, 'imagenet_classifier.jsonnet')
            module2 = StandardModule(imagenet_jsonnet, cfg=cfg)
            module2.eval()
            module2.state_dict_prop = {'classifier.'+k: v for k, v in module.state_dict().items()}
            logits2 = module2(image=image)
            self.assertTrue(torch.all(logits.eq(logits2)))

            torchvision = [
                {
                    'pth': 'resnet18-5c106cde.pth',
                    'class': 'torchvision.models.resnet.resnet18',
                    'num_blocks': [2, 2, 2, 2],
                },
                {
                    'pth': 'resnet34-333f7ec4.pth',
                    'class': 'torchvision.models.resnet.resnet34',
                    'num_blocks': [3, 4, 6, 3],
                },
            ]
            for num in range(len(torchvision)):
                state_dict_path = os.path.join(data_dir, torchvision[num]['pth'])
                if os.path.isfile(state_dict_path):
                    cfg['ext_vars']['num_blocks'] = torchvision[num]['num_blocks']
                    module = StandardModule(resnet_jsonnet, cfg=cfg, state_dict=state_dict_path)
                    module.eval()
                    logits = module(image=image)
                    torchvision_class = import_object(torchvision[num]['class'])
                    module2 = torchvision_class()
                    module2.eval()
                    module2.load_state_dict(torch.load(state_dict_path))
                    logits2 = module2(image)
                    self.assertTrue(torch.all(logits.eq(logits2)))


    def test_squeezenet(self):
        with torch.no_grad():
            module = StandardModule(squeezenet_jsonnet)
            module.eval()
            image = torch.rand(1, 3, 256, 256)
            logits = module(image=image)
            self.assertEqual(list(logits.shape), [1, 1000])


    def test_text_image_classification(self):
        with torch.no_grad():
            module = StandardModule(text_image_jsonnet, cfg=text_image_cfg)
            module.eval()
            text = torch.randint(0, 100, (1, 512))
            image = torch.rand(1, 3, 256, 256)
            logits = module(text=text, image=image)
            self.assertEqual([list(p.shape) for p in logits], [[1, 16], [1, 3]])


    def test_laia(self):
        tmpdir = tempfile.mkdtemp(prefix='_narchi_test_')
        state_dict_path = os.path.join(tmpdir, 'laia.pth')

        with torch.no_grad():
            # Check instantiation and forward #
            module = StandardModule(laia_jsonnet, cfg=laia_cfg)
            module.eval()
            image = torch.rand(1, 3, 64, 128)
            logits = module(image=image)
            self.assertEqual(len(logits.shape), 3)
            self.assertEqual(logits.shape[0], 1)
            self.assertEqual(logits.shape[1], 128/8)
            self.assertEqual(logits.shape[2], laia_ext_vars['num_symbols'])

            # Check forward of reloaded parameters gives same result #
            state_dict = module.state_dict_prop
            torch.save(state_dict, state_dict_path)
            module2 = StandardModule(laia_jsonnet, cfg=laia_cfg, state_dict=state_dict_path)
            module2.eval()
            state_dict2 = module2.state_dict_prop
            self.assertEqual(state_dict.keys(), state_dict2.keys())
            for key in state_dict.keys():
                comp = torch.all(state_dict[key].eq(state_dict2[key]))
                self.assertTrue(comp, 'Reloaded state_dict differs for '+key+'.')
            logits2 = module2(image=image)
            comp = torch.all(logits.eq(logits2))
            self.assertTrue(comp, 'Output from module reloading state_dict differs.')

            # Check invalid instantiations #
            self.assertRaises(RuntimeError, lambda: module(image))
            self.assertRaises(RuntimeError, lambda: module(value=image))
            self.assertRaises(RuntimeError, lambda: module(image=torch.rand(64, 128)))

        shutil.rmtree(tmpdir)


    def test_packed_blocks(self):
        widths = [128, 96, 64]
        images = [torch.rand(3, 64, widths[0]), torch.rand(3, 64, widths[1]), torch.rand(3, 64, widths[2])]

        # Conv2dPacked #
        block = Conv2dPacked(in_channels=3, out_channels=5, kernel_size=7, padding=3)
        packed_images = pack_2d_sequences(images, gap_size=1, length_fact=1)
        self.assertRaises(RuntimeError, lambda: block(packed_images))
        packed_images = pack_2d_sequences(images, gap_size=3, length_fact=1)
        block(packed_images)
        self.assertRaises(NotImplementedError, lambda: Conv2dPacked(in_channels=3, out_channels=5, kernel_size=7, padding=0))

        # MaxPool2dPacked #
        block = MaxPool2dPacked(kernel_size=4, stride=4)
        packed_images = pack_2d_sequences(images, gap_size=1, length_fact=1)
        self.assertRaises(RuntimeError, lambda: block(packed_images))
        packed_images = pack_2d_sequences(images, gap_size=4, length_fact=4)
        block(packed_images)
        self.assertRaises(NotImplementedError, lambda: MaxPool2dPacked(kernel_size=4, stride=2))


    def test_laia_packed(self):
        with torch.no_grad():
            # Check standard and packed give same result #
            module = PackedModule(laia_jsonnet, cfg=laia_cfg, gap_size=4, length_fact=8)
            module.eval()
            module2 = StandardModule(laia_jsonnet, cfg=laia_cfg, state_dict=module.state_dict())
            module2.eval()
            #widths = [129, 91, 65]
            widths = [128, 96, 64]
            images = [torch.rand(3, 64, widths[0]), torch.rand(3, 64, widths[1]), torch.rand(3, 64, widths[2])]
            logits_packed = module(image=images)
            logits, lengths = pad_packed_sequence(logits_packed, batch_first=True)

            for num, image in enumerate(images):
                with self.subTest(f'image {num}'):
                    logits2 = module2(image=image.unsqueeze(0))
                    self.assertEqual(lengths[num], logits2.shape[1])
                    self.assertTrue(torch.allclose(logits2[0,:,:], logits[num,:lengths[num],:], atol=1e-7))

            # Check instantiation from object #
            module2 = ModuleArchitecture(laia_jsonnet, cfg=laia_cfg)
            module2 = PackedModule(module2, gap_size=4, length_fact=8, state_dict=module.state_dict())
            module2.eval()
            logits_packed2 = module2(image=images)
            self.assertTrue(torch.allclose(logits_packed.data, logits_packed2.data))

            # Check debug mode #
            self.assertFalse(hasattr(module, 'intermediate_outputs'))
            module = PackedModule(laia_jsonnet, cfg=laia_cfg, gap_size=4, length_fact=8, debug=True)
            module.eval()
            module(image=images)
            self.assertTrue(hasattr(module, 'intermediate_outputs'))
            intermediate_outputs = list(module.intermediate_outputs.keys())
            self.assertEqual(intermediate_outputs, ['conv1', 'conv2', 'conv3', 'conv4', 'to_1d', 's3blstm', 'fc'])

            # Check invalid gaps #
            module2 = PackedModule(laia_jsonnet, cfg=laia_cfg, gap_size=4, length_fact=4)
            self.assertRaises(RuntimeError, lambda: module2(image=images))

            # Check custom module #
            class TestModule(BaseModule):
                blocks_mappings = packed_pytorch_blocks_mappings

            module2 = TestModule(laia_jsonnet, cfg=laia_cfg)
            module2.eval()
            module2(image=torch.rand(1, 3, 64, 96))

            # Check unsorted images #
            unsorted_images = [images[n] for n in [1, 2, 0]]
            packed = pack_2d_sequences(unsorted_images, fail_if_unsorted=False)
            self.assertEqual(list(packed.lengths), widths)
            self.assertRaises(ValueError, lambda: module(image=unsorted_images))


if __name__ == '__main__':
    unittest.main(verbosity=2)



#!/usr/bin/env python3
"""Unit tests for symbolic operations."""

import os
import unittest
from narchi.sympy import variable_operate, prod


class SympyTests(unittest.TestCase):
    """Tests for symbolic operation functions."""

    def test_variable_operate(self):
        ## successes ##
        examples = [
            {
                'input': 12,
                'operation': '__input__/3',
                'expected': 4,
            },
            {
                'input': '<<variable:X>>',
                'operation': 16,
                'expected': 16,
            },
            {
                'input': '<<variable:X>>',
                'operation': '<<variable:T>>',
                'expected': '<<variable:T>>',
            },
            {
                'input': '<<variable:X>>',
                'operation': '__input__/5',
                'expected': '<<variable:X/5>>',
            },
            {
                'input': '<<variable:X>>',
                'operation': '(__input__+7)+2*(__input__+2)',
                'expected': '<<variable:3*X+11>>',
            },
            {
                'input': '<<variable:8*X+4>>',
                'operation': '__input__/2',
                'expected': '<<variable:4*X+2>>',
            },
            {
                'input': '<<variable:X/12+7*Y>>',
                'operation': '4*__input__',
                'expected': '<<variable:X/3+28*Y>>',
            },
            {
                'input': '<<variable:X>>',
                'operation': '2*__input__+Y',
                'expected': '<<variable:2*X+Y>>',
            },
        ]
        for example in examples:
            result = variable_operate(example['input'], example['operation'])
            self.assertEqual(result, example['expected'])

        ## failures ##
        examples = [
            {
                'input': '<<X>>',
                'operation': '__input__/3',
            },
            {
                'input': '<<variable:__input__>>',
                'operation': '__input__+5',
            },
            {
                'input': '<<variable:X>>',
                'operation': 3.25,
            },
            {
                'input': 7.6,
                'operation': '2*__input__',
            },
        ]
        for example in examples:
            self.assertRaises(ValueError, lambda: variable_operate(example['input'], example['operation']))


    def test_prod(self):
        ## successes ##
        examples = [
            {
                'values': [2, 3, 4],
                'expected': 24,
            },
            {
                'values': ['<<variable:X>>', '<<variable:2*Y>>', 3],
                'expected': '<<variable:6*X*Y>>',
            },
        ]
        for example in examples:
            result = prod(example['values'])
            self.assertEqual(result, example['expected'])

        ## failures ##
        examples = [
            {
                'values': [],
            },
            {
                'values': ['<<variable:X>>', 2.5],
            },
        ]
        for example in examples:
            self.assertRaises(ValueError, lambda: prod(example['values']))


if __name__ == '__main__':
    unittest.main(verbosity=2)



#!/usr/bin/env python3
"""Unit tests for modules."""

import os
import shutil
import tempfile
import unittest
from jsonargparse import ParserError
from jsonschema.exceptions import ValidationError
from narchi.module import ModuleArchitecture
from narchi.blocks import propagators
from narchi.schemas import auto_tag
from narchi_tests.data import *


class ModuleArchitectureTests(unittest.TestCase):
    """Tests for the ModuleArchitecture class."""

    def setUp(self):
        self.tmpdir = tempfile.mkdtemp(prefix='_narchi_test_')
        self.laia_out_json = os.path.join(self.tmpdir, 'laia.json')


    def tearDown(self):
        shutil.rmtree(self.tmpdir)


    def test_failed_class_init(self):
        self.assertRaises(ValueError, lambda: ModuleArchitecture({}, cfg=laia_cfg))
        self.assertRaises(ParserError, lambda: ModuleArchitecture(laia_jsonnet, cfg=None))
        self.assertRaises(ValueError, lambda: ModuleArchitecture(laia_jsonnet, cfg=False))
        self.assertRaises(TypeError, lambda: ModuleArchitecture(laia_jsonnet, cfg='/tmp'))


    def test_propagation_success_failure(self):
        cfg = dict(laia_cfg)
        cfg['propagate'] = False

        module = ModuleArchitecture(laia_jsonnet, cfg=cfg)
        module.propagate()
        self.assertEqual(laia_shapes, [b._shape.out for b in module.architecture.blocks])

        module = ModuleArchitecture(laia_jsonnet, cfg=cfg)
        module.architecture.outputs[0]._shape[0] = auto_tag
        module.propagate()
        self.assertEqual('laia', module.architecture._id)
        self.assertEqual(laia_shapes, [b._shape.out for b in module.architecture.blocks])
        self.assertRaises(RuntimeError, lambda: module.propagate())

        orig_jsonnet = module.jsonnet
        tmp_laia_jsonnet = os.path.join(self.tmpdir, 'laia.jsonnet')
        with open(tmp_laia_jsonnet, 'w') as f:
            f.write(orig_jsonnet.replace('fc -> logits', 'logits -> fc'))
        module = ModuleArchitecture(tmp_laia_jsonnet, cfg=cfg)
        self.assertRaises(ValueError, lambda: module.propagate())

        with open(tmp_laia_jsonnet, 'w') as f:
            f.write(orig_jsonnet.replace('fc -> logits', 'fc').replace('logits', 'fc'))
        self.assertRaises(ValueError, lambda: ModuleArchitecture(tmp_laia_jsonnet, cfg=cfg))

        module = ModuleArchitecture(laia_jsonnet, cfg=cfg)
        module.architecture.outputs[0]._shape[0] = '<<variable:W/16>>'
        self.assertRaises(ValueError, lambda: module.propagate())


    def test_validation_fail(self):
        cfg = dict(laia_cfg)
        cfg['propagate'] = False
        module = ModuleArchitecture(laia_jsonnet, cfg=cfg)
        module.architecture.graph[0] = 'image -- logits'
        self.assertRaises(ValidationError, lambda: module.validate())


    def test_save_json_write_overwrite(self):
        cfg = dict(laia_cfg)
        cfg.update({'outdir': self.tmpdir, 'save_json': True})
        ModuleArchitecture(laia_jsonnet, cfg=cfg)
        self.assertTrue(os.path.isfile(self.laia_out_json))
        self.assertRaises(IOError, lambda: ModuleArchitecture(laia_jsonnet, cfg=cfg))
        cfg['overwrite'] = True
        ModuleArchitecture(laia_jsonnet, cfg=cfg)
        os.remove(self.laia_out_json)


    def test_custom_block_undefined_defined(self):
        cfg = dict(laia_cfg)
        cfg.update({'outdir': self.tmpdir, 'save_json': True, 'propagate': False})
        module = ModuleArchitecture(laia_jsonnet, cfg=cfg)
        custom_block = module.architecture.blocks[0]
        custom_propagators = dict(propagators)
        custom_propagators['Custom'] = propagators[custom_block._class]
        custom_block._class = 'Custom'
        self.assertRaises(ValueError, lambda: module.propagate())
        self.assertTrue(os.path.isfile(self.laia_out_json))
        os.remove(self.laia_out_json)
        cfg.update({'save_json': False, 'propagators': custom_propagators})
        ModuleArchitecture(laia_jsonnet, cfg=cfg).propagate()


    def test_multi_input_output(self):
        cfg = dict(text_image_cfg)

        module = ModuleArchitecture(text_image_jsonnet, cfg=cfg)
        self.assertEqual([i._shape for i in module.architecture.inputs], [['<<variable:L>>'], [3, '<<variable:H>>', '<<variable:W>>']])
        self.assertEqual([o._shape for o in module.architecture.outputs], [[16], [3]])

        del cfg['ext_vars']
        self.assertRaises(ParserError, lambda: ModuleArchitecture(text_image_jsonnet, cfg=cfg))


    def test_nested_modules(self):
        cfg = {'ext_vars': nested3_ext_vars}
        module = ModuleArchitecture(nested3_jsonnet, cfg=cfg)
        self.assertEqual({'in': [64], 'out': [16]}, vars(module.architecture._shape))

        cfg = {'ext_vars': nested2_ext_vars}
        module = ModuleArchitecture(nested2_jsonnet, cfg=cfg)
        self.assertEqual({'in': [128], 'out': [16]}, vars(module.architecture._shape))

        cfg = {'ext_vars': nested1_ext_vars}
        module = ModuleArchitecture(nested1_jsonnet, cfg=cfg)
        self.assertEqual({'in': [128], 'out': [16]}, vars(module.architecture._shape))


if __name__ == '__main__':
    unittest.main(verbosity=2)



#!/usr/bin/env python3
"""Unit tests for propagation classes."""

import unittest
from copy import deepcopy
from jsonargparse import dict_to_namespace as d2n
from narchi.blocks import propagators, register_propagator
from narchi.schemas import auto_tag
from narchi.propagators.base import BasePropagator, get_shape, create_shape
from narchi.propagators.concat import ConcatenatePropagator
from narchi.propagators.conv import ConvPropagator, PoolPropagator
from narchi.propagators.fixed import AddFixedPropagator, FixedOutputPropagator
from narchi.propagators.group import SequentialPropagator
from narchi.propagators.reshape import ReshapePropagator
from narchi.propagators.rnn import RnnPropagator
from narchi.propagators.same import SameShapePropagator, SameShapesPropagator, SameShapeConsumeDimPropagator
from narchi.graph import parse_graph


class BasePropagatorTests(unittest.TestCase):
    """Tests for the BasePropagator class."""

    def test_propagators_dict(self):
        self.assertRaises(ValueError, lambda: register_propagator(FixedOutputPropagator('Linear')))
        self.assertRaises(ValueError, lambda: register_propagator(FixedOutputPropagator('Default')))
        register_propagator(FixedOutputPropagator('Linear'), replace=True)
        for propagator in propagators.values():
            self.assertTrue(isinstance(propagator, BasePropagator))


    def test_base_propagator(self):
        propagator = BasePropagator('Base')

        examples = [
            {
                'from': [{}],
                'to': d2n({'_id': 'b2', '_class': 'Base'}),
            },
            {
                'from': [d2n({'_id': 'b1'})],
                'to': d2n({'_id': 'b2', '_class': 'Base'}),
            },
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': [5]}})],
                'to': d2n({}),
            },
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': [5]}})],
                'to': d2n({'_id': 'b2', '_class': 'Base2'}),
            },
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': [5]}})],
                'to': d2n({'_id': 'b2', '_class': 'Base', '_shape': {'in': [5], 'out': [5]}}),
            },
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': [auto_tag]}})],
                'to': d2n({'_id': 'b2', '_class': 'Base'}),
            },
        ]
        for example in examples:
            self.assertRaises(ValueError, lambda: propagator.initial_checks(example['from'], example['to']))

        self.assertRaises(NotImplementedError, lambda: propagator.propagate(None, None))


    def test_bad_subclass(self):
        class BadPropagator(BasePropagator):
            def propagate(self, from_blocks, block):
                shape_in = get_shape('out', from_blocks[0])
                if any(isinstance(x, int) for x in shape_in):
                    block._shape = create_shape(shape_in, [auto_tag])
                else:
                    block._shape = create_shape(['<<variable:V>>'], shape_in)

        propagator = BadPropagator('Bad')

        examples = [
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': [16]}})],
                'to': d2n({'_id': 'b2', '_class': 'Bad'}),
            },
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': ['<<variable:Z>>']}})],
                'to': d2n({'_id': 'b2', '_class': 'Bad'}),
            },
        ]
        for example in examples:
            self.assertRaises(ValueError, lambda: propagator(example['from'], example['to']))


class AddFixedPropagatorTests(unittest.TestCase):
    """Tests for the AddFixedPropagator class."""

    def test_initializer(self):
        self.assertRaises(ValueError, lambda: AddFixedPropagator('Test', fixed_dims=0))


    def test_embedding(self):
        propagator = propagators['Embedding']

        ## successes ##
        examples = [
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': [21]}})],
                'to': d2n({'_id': 'b2', '_class': 'Embedding',
                           'output_feats': 17}),
                'expected': [21, 17],
            },
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': [3, '<<variable:L>>']}})],
                'to': d2n({'_id': 'b2', '_class': 'Embedding',
                           'output_feats': 25}),
                'expected': [3, '<<variable:L>>', 25],
            },
        ]
        for example in examples:
            from_block = example['from']
            block = example['to']
            propagator(from_block, block)
            self.assertEqual(block._shape.out, example['expected'])


class FixedOutputPropagatorTests(unittest.TestCase):
    """Tests for the FixedOutputPropagator class."""

    def test_initializer(self):
        self.assertRaises(ValueError, lambda: FixedOutputPropagator('Test', fixed_dims=0))
        self.assertRaises(ValueError, lambda: FixedOutputPropagator('Test', unfixed_dims=0))

        propagator = FixedOutputPropagator('Fixed', fixed_dims=2)
        example = {
            'from': [d2n({'_id': 'b1', '_shape': {'out': [480]}})],
            'to': d2n({'_id': 'b2', '_class': 'Fixed',
                       'output_feats': [1, 480]}),
        }
        self.assertRaises(ValueError, lambda: propagator(example['from'], example['to']))


    def test_linear_propagations(self):
        propagator = FixedOutputPropagator('Linear')

        ## successes ##
        examples = [
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': [21]}})],
                'to': d2n({'_id': 'b2', '_class': 'Linear',
                           'output_feats': 17}),
                'expected': [17],
            },
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': [7, 16]}})],
                'to': d2n({'_id': 'b2', '_class': 'Linear',
                           'output_feats': 9}),
                'expected': [7, 9],
            },
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': [3, '<<variable:2*X>>', 24, '<<variable:Y>>']}})],
                'to': d2n({'_id': 'b2', '_class': 'Linear',
                           'output_feats': 6}),
                'expected': [3, '<<variable:2*X>>', 24, 6],
            },
        ]
        for example in examples:
            from_block = example['from']
            block = example['to']
            propagator(from_block, block)
            self.assertEqual(block._shape.out, example['expected'])

        ## failures ##
        examples = [
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': [7, 16]}})],
                'to': d2n({'_id': 'b2', '_class': 'Linear'}),
            },
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': []}})],
                'to': d2n({'_id': 'b2', '_class': 'Linear',
                           'output_feats': 17}),
            },
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': [7, 16]}}),
                         d2n({'_id': 'b2', '_shape': {'out': [7, 16]}})],
                'to': d2n({'_id': 'b3', '_class': 'Linear',
                           'output_feats': 17}),
            },
        ]
        for example in examples:
            self.assertRaises(ValueError, lambda: propagator(example['from'], example['to']))


    def test_adaptiveavgpool_propagations(self):
        propagator = propagators['AdaptiveAvgPool2d']

        ## successes ##
        examples = [
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': [3, 480, 640]}})],
                'to': d2n({'_id': 'b2', '_class': 'AdaptiveAvgPool2d',
                           'output_feats': [1, 1]}),
                'expected': [3, 1, 1],
            },
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': ['<<variable:C>>', '<<variable:H>>', '<<variable:W>>']}})],
                'to': d2n({'_id': 'b2', '_class': 'AdaptiveAvgPool2d',
                           'output_feats': ['<<variable:VG>>', '<<variable:HG>>']}),
                'expected': ['<<variable:C>>', '<<variable:VG>>', '<<variable:HG>>'],
            },
        ]
        for example in examples:
            from_block = example['from']
            block = example['to']
            propagator(from_block, block)
            self.assertEqual(block._shape.out, example['expected'])

        ## failures ##
        examples = [
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': [3, 480, 640]}})],
                'to': d2n({'_id': 'b2', '_class': 'AdaptiveAvgPool2d', 'output_feats': [0, 0]}),
            },
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': ['<<variable:H>>', '<<variable:W>>']}})],
                'to': d2n({'_id': 'b2', '_class': 'AdaptiveAvgPool2d', 'output_feats': ['<<variable:VG>>', '<<variable:HG>>']}),
            },
        ]
        for example in examples:
            self.assertRaises(ValueError, lambda: propagator(example['from'], example['to']))


class SameShapePropagatorsTests(unittest.TestCase):
    """Tests for propagator classes that preserve shape."""

    def test_same_shape_propagations(self):
        propagator = SameShapePropagator('SameShape')

        ## successes ##
        examples = [
            [5],
            ['<<variable:A>>', 3],
            [7, 3, '<<variable:A>>'],
        ]
        for example in examples:
            from_block = d2n({'_id': 'b1', '_shape': {'out': example}})
            block = d2n({'_id': 'b2', '_class': 'SameShape'})
            propagator([from_block], block)
            self.assertEqual(from_block._shape.out, block._shape.out)
            self.assertEqual(vars(block._shape)['in'], block._shape.out)

        ## failures ##
        examples = [
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': [5]}})],
                'to': d2n({'_id': 'b2', '_class': 'SameShape2'}),
            },
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': auto_tag}})],
                'to': d2n({'_id': 'b2', '_class': 'SameShape'}),
            },
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': [5]}}),
                         d2n({'_id': 'b2', '_shape': {'out': [7]}})],
                'to': d2n({'_id': 'b3', '_class': 'SameShape'}),
            },
        ]
        for example in examples:
            self.assertRaises(ValueError, lambda: propagator(example['from'], example['to']))


    def test_same_shapes_propagations(self):
        propagator = SameShapesPropagator('SameShapes')

        ## successes ##
        examples = [
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': ['<<variable:4*X>>']}}),
                         d2n({'_id': 'b2', '_shape': {'out': ['<<variable:4*X>>']}})],
                'to': d2n({'_id': 'b3', '_class': 'SameShapes'}),
            },
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': [26, '<<variable:X>>']}}),
                         d2n({'_id': 'b2', '_shape': {'out': [26, '<<variable:X>>']}}),
                         d2n({'_id': 'b3', '_shape': {'out': [26, '<<variable:X>>']}})],
                'to': d2n({'_id': 'b4', '_class': 'SameShapes'}),
            },
        ]
        for example in examples:
            propagator(example['from'], example['to'])
            for from_block in example['from']:
                self.assertEqual(from_block._shape.out, example['to']._shape.out)
            self.assertEqual(vars(example['to']._shape)['in'], example['to']._shape.out)

        ## failures ##
        examples = [
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': [5]}})],
                'to': d2n({'_id': 'b3', '_class': 'SameShapes'}),
            },
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': ['<<variable:4*X>>']}}),
                         d2n({'_id': 'b2', '_shape': {'out': ['<<variable:X>>']}})],
                'to': d2n({'_id': 'b3', '_class': 'SameShapes'}),
            },
        ]
        for example in examples:
            self.assertRaises(ValueError, lambda: propagator(example['from'], example['to']))

    def test_same_shape_consume_dim_propagations(self):
        propagator = propagators['CRF']

        ## successes ##
        examples = [
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': [7, 5]}})],
                'to': d2n({'_id': 'b2', '_class': 'CRF'}),
                'expected': [7],
            },
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': ['<<variable:L>>', 3]}})],
                'to': d2n({'_id': 'b2', '_class': 'CRF'}),
                'expected': ['<<variable:L>>'],
            },
        ]
        for example in examples:
            from_blocks = example['from']
            block = example['to']
            propagator(from_blocks, block)
            self.assertEqual(block._shape.out, example['expected'])

        ## failures ##
        examples = [
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': [7]}})],
                'to': d2n({'_id': 'b2', '_class': 'CRF'}),
            },
        ]
        for example in examples:
            self.assertRaises(ValueError, lambda: propagator(example['from'], example['to']))


class ConcatPropagatorsTests(unittest.TestCase):
    """Tests for the concatenation propagation classes."""

    def test_concatenate_propagations(self):
        propagator = propagators['Concatenate']

        ## successes ##
        examples = [
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': [3, 48, 32]}}),
                         d2n({'_id': 'b2', '_shape': {'out': [4, 48, 32]}}),
                         d2n({'_id': 'b3', '_shape': {'out': [5, 48, 32]}})],
                'to': d2n({'_id': 'b4', '_class': 'Concatenate', 'dim': 0}),
                'expected': [12, 48, 32],
            },
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': ['<<variable:C>>', '<<variable:H>>', '<<variable:W>>']}}),
                         d2n({'_id': 'b2', '_shape': {'out': ['<<variable:C+2>>', '<<variable:H>>', '<<variable:W>>']}})],
                'to': d2n({'_id': 'b4', '_class': 'Concatenate', 'dim': -3}),
                'expected': ['<<variable:2*C+2>>', '<<variable:H>>', '<<variable:W>>'],
            },
        ]
        for example in examples:
            from_blocks = example['from']
            block = example['to']
            propagator(from_blocks, block)
            self.assertEqual(block._shape.out, example['expected'])

        ## failures ##
        examples = [
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': [3, 48, 32]}}),
                         d2n({'_id': 'b2', '_shape': {'out': [4, 48, 32]}}),
                         d2n({'_id': 'b3', '_shape': {'out': [5, 48, 32]}})],
                'to': d2n({'_id': 'b4', '_class': 'Concatenate'}),
            },
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': [3, 48, 32]}})],
                'to': d2n({'_id': 'b4', '_class': 'Concatenate', 'dim': 0}),
            },
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': [3, 48, 32]}}),
                         d2n({'_id': 'b2', '_shape': {'out': [4, 32, 48]}}),
                         d2n({'_id': 'b3', '_shape': {'out': [5, 48, 32]}})],
                'to': d2n({'_id': 'b4', '_class': 'Concatenate', 'dim': 0}),
            },
        ]
        for example in examples:
            self.assertRaises(ValueError, lambda: propagator(example['from'], example['to']))


class ConvPropagatorTests(unittest.TestCase):
    """Tests for the ConvPropagator class."""

    def test_init(self):
        self.assertRaises(ValueError, lambda: ConvPropagator('Bad', conv_dims=0))


    def test_bad_subclass(self):
        class BadPropagator(ConvPropagator):
            num_features_source = 'out_features'
        self.assertRaises(ValueError, lambda: BadPropagator('Bad', conv_dims=1))


    def test_conv_1d_propagations(self):
        propagator = ConvPropagator('Conv1d', conv_dims=1)

        ## successes ##
        examples = [
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': [7, 16]}})],
                'to': d2n({'_id': 'b2', '_class': 'Conv1d',
                           'output_feats': 9, 'kernel_size': 3, 'padding': 1}),
                'expected': [9, 16],
            },
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': [3, '<<variable:L>>']}})],
                'to': d2n({'_id': 'b2', '_class': 'Conv1d',
                           'output_feats': 6, 'kernel_size': 5, 'padding': 2}),
                'expected': [6, '<<variable:L>>'],
            },
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': [5, '<<variable:L>>']}})],
                'to': d2n({'_id': 'b2', '_class': 'Conv1d',
                           'output_feats': 2, 'kernel_size': 2, 'stride': 2}),
                'expected': [2, '<<variable:L/2>>'],
            },
        ]
        for example in examples:
            from_block = example['from']
            block = example['to']
            propagator(from_block, block)
            self.assertEqual(block._shape.out, example['expected'])

        ## failures ##
        examples = [
            {
                'from': [d2n({'_id': 'b1'})],
                'to': d2n({'_id': 'b2', '_class': 'Conv1d',
                           'output_feats': 9, 'kernel_size': 3, 'padding': 1}),
            },
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': [7]}})],
                'to': d2n({'_id': 'b2', '_class': 'Conv1d',
                           'output_feats': 9, 'kernel_size': 3, 'padding': 1}),
            },
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': [3, '<<variable:L>>', 16]}})],
                'to': d2n({'_id': 'b2', '_class': 'Conv1d',
                           'output_feats': 6, 'kernel_size': 5, 'padding': 2}),
            },
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': [5, '<<L>>']}})],
                'to': d2n({'_id': 'b2', '_class': 'Conv1d',
                           'output_feats': 2, 'kernel_size': 2, 'stride': 2}),
            },
        ]
        for example in examples:
            self.assertRaises(ValueError, lambda: propagator(example['from'], example['to']))


    def test_conv_2d_propagations(self):
        propagator = ConvPropagator('Conv2d', conv_dims=2)

        ## successes ##
        examples = [
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': [7, 16, 128]}})],
                'to': d2n({'_id': 'b2', '_class': 'Conv2d',
                           'output_feats': 9, 'kernel_size': 3, 'padding': 1}),
                'expected': [9, 16, 128],
            },
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': [3, 24, '<<variable:W>>']}})],
                'to': d2n({'_id': 'b2', '_class': 'Conv2d',
                           'output_feats': 6, 'kernel_size': 5, 'padding': 2}),
                'expected': [6, 24, '<<variable:W>>'],
            },
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': [5, '<<variable:H>>', '<<variable:W>>']}})],
                'to': d2n({'_id': 'b2', '_class': 'Conv2d',
                           'output_feats': 2, 'kernel_size': 3, 'stride': 3}),
                'expected': [2, '<<variable:H/3>>', '<<variable:W/3>>'],
            },
        ]
        for example in examples:
            from_block = example['from']
            block = example['to']
            propagator(from_block, block)
            self.assertEqual(block._shape.out, example['expected'])

        ## failures ##
        examples = [
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': [7, 16, 128]}})],
                'to': d2n({'_id': 'b2', '_class': 'Conv2d',
                           'output_feats': 0, 'kernel_size': 3, 'padding': 1}),
            },
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': [7, 16, 128]}})],
                'to': d2n({'_id': 'b2', '_class': 'Conv2d',
                           'kernel_size': 3, 'padding': 1}),
            },
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': [32, 3, 24, '<<variable:W>>']}})],
                'to': d2n({'_id': 'b2', '_class': 'Conv2d',
                           'output_feats': 6, 'kernel_size': 5, 'padding': 2}),
            },
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': [5]}})],
                'to': d2n({'_id': 'b2', '_class': 'Conv2d',
                           'output_feats': 2, 'kernel_size': 3, 'stride': 3}),
            },
        ]
        for example in examples:
            self.assertRaises(ValueError, lambda: propagator(example['from'], example['to']))


class PoolPropagatorTests(unittest.TestCase):
    """Tests for the PoolPropagator class."""

    def test_pool_1d_propagations(self):
        propagator = PoolPropagator('Pool1d', conv_dims=1)
        examples = [
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': [7, 16]}})],
                'to': d2n({'_id': 'b2', '_class': 'Pool1d',
                           'kernel_size': 3, 'padding': 1}),
                'expected': [7, 16],
            },
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': [3, '<<variable:L>>']}})],
                'to': d2n({'_id': 'b2', '_class': 'Pool1d',
                           'kernel_size': 5, 'padding': 2}),
                'expected': [3, '<<variable:L>>'],
            },
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': [5, '<<variable:L>>']}})],
                'to': d2n({'_id': 'b2', '_class': 'Pool1d',
                           'kernel_size': 2, 'stride': 2}),
                'expected': [5, '<<variable:L/2>>'],
            },
        ]
        for example in examples:
            from_block = example['from']
            block = example['to']
            propagator(from_block, block)
            self.assertEqual(block._shape.out, example['expected'])

    def test_pool_2d_propagations(self):
        propagator = PoolPropagator('Pool2d', conv_dims=2)
        examples = [
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': [7, 16, 128]}})],
                'to': d2n({'_id': 'b2', '_class': 'Pool2d',
                           'kernel_size': 3, 'padding': 1}),
                'expected': [7, 16, 128],
            },
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': [3, 24, '<<variable:W>>']}})],
                'to': d2n({'_id': 'b2', '_class': 'Pool2d',
                           'kernel_size': 5, 'padding': 2}),
                'expected': [3, 24, '<<variable:W>>'],
            },
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': ['<<variable:C>>', '<<variable:H>>', '<<variable:W>>']}})],
                'to': d2n({'_id': 'b2', '_class': 'Pool2d',
                           'kernel_size': 3, 'stride': 3}),
                'expected': ['<<variable:C>>', '<<variable:H/3>>', '<<variable:W/3>>'],
            },
        ]
        for example in examples:
            from_block = example['from']
            block = example['to']
            propagator(from_block, block)
            self.assertEqual(block._shape.out, example['expected'])


class RnnPropagatorTests(unittest.TestCase):
    """Tests for the RnnPropagator class."""

    def test_rnn_propagations(self):
        propagator = RnnPropagator('RNN')

        ## successes ##
        examples = [
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': [128, 16]}})],
                'to': d2n({'_id': 'b2', '_class': 'RNN',
                           'output_feats': 8}),
                'expected': [128, 8],
            },
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': ['<<variable:L>>', 14]}})],
                'to': d2n({'_id': 'b2', '_class': 'RNN',
                           'output_feats': 12, 'bidirectional': True}),
                'expected': ['<<variable:L>>', 12],
            },
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': ['<<variable:L>>', '<<variable:F>>']}})],
                'to': d2n({'_id': 'b2', '_class': 'RNN',
                           'output_feats': 7}),
                'expected': ['<<variable:L>>', 7],
            },
        ]
        for example in examples:
            from_block = example['from']
            block = example['to']
            propagator(from_block, block)
            self.assertEqual(block._shape.out, example['expected'])
            self.assertEqual(block.output_feats/(2 if block.bidirectional else 1), block.hidden_size)

        ## failures ##
        examples = [
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': [3, 128, 16]}})],
                'to': d2n({'_id': 'b2', '_class': 'RNN',
                           'output_feats': 8}),
            },
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': ['<<variable:L>>', 14]}})],
                'to': d2n({'_id': 'b2', '_class': 'RNN',
                           'bidirectional': True}),
                'expected': ['<<variable:L>>', 12],
            },
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': ['<<variable:L>>', 14]}})],
                'to': d2n({'_id': 'b2', '_class': 'RNN',
                           'output_feats': 11, 'bidirectional': True}),
            },
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': ['<<variable:L>>', '<<variable:F>>']}})],
                'to': d2n({'_id': 'b2', '_class': 'RNN',
                           'output_feats': '7'}),
            },
        ]
        for example in examples:
            self.assertRaises(ValueError, lambda: propagator(example['from'], example['to']))


class ReshapePropagatorTests(unittest.TestCase):
    """Tests for the ReshapePropagator class."""

    def test_reshape_propagations(self):
        propagator = ReshapePropagator('Reshape')

        ## successes ##
        examples = [
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': [32, 8, '<<variable:W/8>>']}})],
                'to': d2n({'_id': 'b2', '_class': 'Reshape', 'reshape_spec': [2, [0, 1]]}),
                'expected': ['<<variable:W/8>>', 256],
            },
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': ['<<variable:H/2>>', 8, '<<variable:W/8>>', 2]}})],
                'to': d2n({'_id': 'b2', '_class': 'Reshape', 'reshape_spec': [[3, 2], [0, 1]]}),
                'expected': ['<<variable:W/4>>', '<<variable:4*H>>'],
            },
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': [4608]}})],
                'to': d2n({'_id': 'b2', '_class': 'Reshape', 'reshape_spec': [{'0': [3, 48, auto_tag]}]}),
                'expected': [3, 48, 32],
            },
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': ['<<variable:C*H*W>>']}})],
                'to': d2n({'_id': 'b2', '_class': 'Reshape', 'reshape_spec': [{'0': ['<<variable:C>>', auto_tag, '<<variable:W>>']}]}),
                'expected': ['<<variable:C>>', '<<variable:H>>', '<<variable:W>>'],
            },
        ]
        for example in examples:
            from_block = example['from']
            block = example['to']
            propagator(from_block, block)
            self.assertEqual(block._shape.out, example['expected'])

        ## failures ##
        examples = [
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': [32, 8, '<<variable:W/8>>']}})],
                'to': d2n({'_id': 'b2', '_class': 'Reshape'}),
            },
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': [32, 8, '<<variable:W/8>>']}})],
                'to': d2n({'_id': 'b2', '_class': 'Reshape', 'reshape_spec': []}),
            },
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': [10, 32, 8, '<<variable:W/8>>']}})],
                'to': d2n({'_id': 'b2', '_class': 'Reshape', 'reshape_spec': [2, [0, 1]]}),
            },
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': [32, 8, '<<variable:W/8>>']}})],
                'to': d2n({'_id': 'b2', '_class': 'Reshape', 'reshape_spec': [3, [0, 1]]}),
            },
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': [4608]}})],
                'to': d2n({'_id': 'b2', '_class': 'Reshape', 'reshape_spec': [{'0': [3, 49, auto_tag]}]}),
                'expected': [3, 48, 32],
            },
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': [4608]}})],
                'to': d2n({'_id': 'b2', '_class': 'Reshape', 'reshape_spec': [{'0': [3, auto_tag, auto_tag]}]}),
            },
        ]
        for example in examples:
            self.assertRaises(ValueError, lambda: propagator(example['from'], example['to']))


class GroupPropagatorTests(unittest.TestCase):
    """Tests for the group propagation classes."""

    def test_sequential_propagations(self):
        propagator = propagators['Sequential']

        ## successes ##
        examples = [
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': [3, '<<variable:H>>', '<<variable:W>>']}})],
                'to': d2n({'_id': 'b2', '_class': 'Sequential',
                           'blocks': [
                                {'_id': 'i1', '_class': 'Conv2d', 'output_feats': 8, 'kernel_size': 7, 'padding': 3},
                                {'_id': 'i2', '_class': 'BatchNorm2d'},
                                {'_id': 'i3', '_class': 'MaxPool2d', 'kernel_size': 2, 'stride': 2},
                                {'_id': 'i4', '_class': 'Linear', 'output_feats': 12}]}),
                'expected': [8, '<<variable:H/2>>', 12],
            },
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': [3, '<<variable:H>>', '<<variable:W>>']}})],
                'to': d2n({'_id': 'b1', '_class': 'Sequential',
                           'blocks': [
                                {'_id': 'i1', '_class': 'Conv2d', 'output_feats': 8, 'kernel_size': 7, 'padding': 3},
                                {'_id': 'i2', '_class': 'BatchNorm2d'},
                                {'_id': 'i3', '_class': 'Sequential',
                                 'blocks': [
                                      {'_id': 'i4', '_class': 'MaxPool2d', 'kernel_size': 2, 'stride': 2},
                                      {'_id': 'i5', '_class': 'Linear', 'output_feats': 12}]}]}),
                'expected': [8, '<<variable:H/2>>', 12],
            },
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': [3, '<<variable:H>>', '<<variable:W>>']}})],
                'to': d2n({'_id': 'b1', '_class': 'Sequential',
                           'blocks': [
                                {'_class': 'Conv2d', 'output_feats': 8, 'kernel_size': 7, 'padding': 3},
                                {'_class': 'BatchNorm2d'},
                                {'_class': 'Sequential',
                                 'blocks': [
                                      {'_class': 'MaxPool2d', 'kernel_size': 2, 'stride': 2},
                                      {'_class': 'Linear', 'output_feats': 12}]}]}),
                'expected': [8, '<<variable:H/2>>', 12],
            },
        ]
        for example in examples:
            from_block = example['from']
            block = example['to']
            propagator(from_block, block, propagators)
            self.assertEqual(block._shape.out, example['expected'])

        ## failures ##
        examples = [
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': [3, '<<variable:H>>', '<<variable:W>>']}})],
                'to': d2n({'_id': 'b2', '_class': 'Sequential'}),
            },
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': [10, 32, 8, '<<variable:W/8>>']}})],
                'to': d2n({'_id': 'b2', '_class': 'Sequential', 'blocks': []}),
            },
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': [3, 48]}})],
                'to': d2n({'_id': 'b2', '_class': 'Sequential',
                           'blocks': [
                                {'_id': 's1', '_class': 'Linear', 'output_feats': 12},
                                {'_id': 's2', '_class': 'Unregistered', 'output_feats': 5}]}),
            },
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': [3, 48]}})],
                'to': d2n({'_id': 'b2', '_class': 'Sequential',
                           'blocks': [
                                {'_id': 's1', '_class': 'Linear', 'output_feats': 12},
                                {'_id': 's1', '_class': 'Linear', 'output_feats': 5}]}),
            },
        ]
        for example in examples:
            self.assertRaises(ValueError, lambda: propagator(example['from'], example['to'], propagators))


    def test_group_propagations(self):
        propagator = propagators['Group']

        base_example = {
            'from': [d2n({'_id': 'b1', '_shape': {'out': [16, '<<variable:H>>', '<<variable:W>>']}})],
            'to': d2n({'_id': 'b2', '_class': 'Group', 'input': 'in', 'output': 'add',
                       'graph': [
                           'in -> conv -> add',
                           'in -> add'],
                       'blocks': [
                           {'_id': 'in', '_class': 'Identity'},
                           {'_id': 'conv', '_class': 'Conv2d', 'output_feats': 16, 'kernel_size': 3, 'padding': 1},
                           {'_id': 'add', '_class': 'Add'}]}),
            'expected': [16, '<<variable:H>>', '<<variable:W>>'],
        }

        nested_example = deepcopy(base_example)
        nested_example['to'].blocks[1] = deepcopy(base_example['to'])
        nested_example['to'].blocks[1]._id = 'conv'

        ## successes ##
        examples = [
            deepcopy(base_example),
            nested_example,
            {
                'from': [d2n({'_id': 'b1', '_shape': {'out': [16, '<<variable:H>>', '<<variable:W>>']}})],
                'to': d2n({'_id': 'b2', '_class': 'Group', 'input': 'in', 'output': 'relu',
                           'graph': [
                               'in -> conv -> bn -> add -> relu',
                               'in -> add'],
                           'blocks': [
                               {'_id': 'in', '_class': 'Identity'},
                               {'_id': 'conv', '_class': 'Conv2d', 'output_feats': 16, 'kernel_size': 3, 'padding': 1},
                               {'_id': 'bn', '_class': 'BatchNorm2d'},
                               {'_id': 'add', '_class': 'Add'},
                               {'_id': 'relu', '_class': 'ReLU'}]}),
                'expected': [16, '<<variable:H>>', '<<variable:W>>'],
            },
        ]
        for example in examples:
            from_block = example['from']
            block = example['to']
            propagator(from_block, block, propagators)
            self.assertEqual(block._shape.out, example['expected'])

        ## failures ##
        example = deepcopy(base_example)
        delattr(example['to'], 'input')
        self.assertRaises(ValueError, lambda: propagator(example['from'], example['to'], propagators))

        example = deepcopy(base_example)
        delattr(example['to'], 'output')
        self.assertRaises(ValueError, lambda: propagator(example['from'], example['to'], propagators))

        example = deepcopy(base_example)
        delattr(example['to'], 'graph')
        self.assertRaises(ValueError, lambda: propagator(example['from'], example['to'], propagators))

        example = deepcopy(base_example)
        delattr(example['to'], 'blocks')
        self.assertRaises(ValueError, lambda: propagator(example['from'], example['to'], propagators))

        example = deepcopy(base_example)
        example['to'].blocks[1]._id = 'in'
        self.assertRaises(ValueError, lambda: propagator(example['from'], example['to'], propagators))

        example = deepcopy(base_example)
        example['to'].graph[0] = 'conv -> add'
        self.assertRaises(ValueError, lambda: propagator(example['from'], example['to'], propagators))

        example = deepcopy(base_example)
        example['to'].graph[1] = 'in -> unk -> add'
        self.assertRaises(ValueError, lambda: propagator(example['from'], example['to'], propagators))

        example = deepcopy(base_example)
        example['to'].graph[0] = '---'
        self.assertRaises(ValueError, lambda: propagator(example['from'], example['to'], propagators))
        self.assertRaises(ValueError, lambda: parse_graph(example['from'], example['to']))  # ResourceWarning: unclosed file <_io.BufferedRandom name=3> context = None

        example = deepcopy(base_example)
        example['to'].graph[1] = 'add -> in'
        self.assertRaises(ValueError, lambda: propagator(example['from'], example['to'], propagators))


if __name__ == '__main__':
    unittest.main(verbosity=2)



#!/usr/bin/env python3
"""Unit tests for the json schema."""

import json
import unittest
from narchi.schemas import jsonvalidator, block_schema, narchi_schema, schema_as_str


class SchemaTests(unittest.TestCase):
    """Tests for the json schema."""

    def test_narchi_schema(self):
        jsonvalidator.check_schema(narchi_schema)
        self.assertRegex(narchi_schema['$id'], '.*narchi/[0-9]+.[0-9]+/schema.json$')
        self.assertEqual(narchi_schema, json.loads(schema_as_str()))


    def test_block_schema(self):
        jsonvalidator.check_schema(block_schema)


if __name__ == '__main__':
    unittest.main(verbosity=2)



"""Propagator classes for groups of blocks."""

import re
import inspect
from jsonargparse import Namespace
from typing import Dict, List
from .base import BasePropagator, get_shape, create_shape
from ..graph import parse_graph
from ..schemas import id_separator


def get_blocks_dict(blocks: List[dict]) -> Dict[str, dict]:
    """Function that creates a dictionary of blocks using _id as keys.

    Args:
        blocks: List of blocks objects.

    Returns:
        Dictionary of blocks.
    """
    blocks_dict = {}
    for block in blocks:
        if block._id in blocks_dict:
            raise ValueError(f'Duplicate block id: {block._id}.')
        blocks_dict[block._id] = block
    return blocks_dict


def add_ids_prefix(block, io_blocks, skip_io=True):
    """Adds to block id a prefix consisting of parent id and separator as defined in propagated schema."""
    prefix = block._id + id_separator
    for num, subblock in enumerate(block.blocks):
        if hasattr(block, '_class') and block._class == 'Sequential' and not hasattr(subblock, '_id'):
            subblock._id = prefix + str(num)
        else:
            subblock._id = prefix + subblock._id
    if hasattr(block, 'input'):
        block.input = prefix + block.input
    if hasattr(block, 'output'):
        block.output = prefix + block.output
    if hasattr(block, 'inputs') and not skip_io:
        for node in block.inputs:
            node._id = prefix + node._id
    if hasattr(block, 'outputs') and not skip_io:
        for node in block.outputs:
            node._id = prefix + node._id
    if hasattr(block, 'graph'):
        skip_ids = set() if skip_io else {b._id for b in io_blocks}
        re_nodes = re.compile(' +-> +')
        for num, graph_line in enumerate(block.graph):
            nodes = re_nodes.split(graph_line)
            nodes = [n if n in skip_ids else prefix+n for n in nodes]
            block.graph[num] = ' -> '.join(nodes)


def propagate_shapes(
    blocks_dict: Dict[str, dict],
    topological_predecessors: Dict[str, List[str]],
    propagators: dict,
    ext_vars: dict,
    cwd: str,
    skip_ids: set = None,
):
    """Function that propagates shapes in blocks based on a connections mapping.

    Args:
        blocks_dict: Dictionary of blocks.
        topological_predecessors: Mapping of block IDs to its input blocks IDs.
        propagators: Dictionary of propagators.
        ext_vars: Dictionary of external variables required to load jsonnet.
        cwd: Working directory to resolve relative paths.
        skip_ids: Blocks that should be skipped in propagation.

    Raises:
        ValueError: If there graph references an undefined block.
        ValueError: If no propagator found for some block.
    """
    if skip_ids is None:
        skip_ids = set()

    for node_to, nodes_from in topological_predecessors.items():
        if node_to in skip_ids:
            continue
        from_blocks = [blocks_dict[n] for n in nodes_from]
        if node_to not in blocks_dict:
            block_ids = {k for k in blocks_dict.keys()}
            raise ValueError(f'Graph references block[id={node_to}] which is not found among ids={block_ids}.')
        block = blocks_dict[node_to]
        if block._class not in propagators:
            raise ValueError(f'No propagator found for block[id={block._id}] of type {block._class}.')
        propagator = propagators[block._class]
        func_param = {x.name for x in inspect.signature(propagator).parameters.values()}
        kwargs = {}
        if 'propagators' in func_param:
            kwargs['propagators'] = propagators
        if 'ext_vars' in func_param:
            kwargs['ext_vars'] = ext_vars
        if 'cwd' in func_param:
            kwargs['cwd'] = cwd
        propagator(from_blocks, block, **kwargs)

    return blocks_dict


class SequentialPropagator(BasePropagator):
    """Propagator for a sequence of blocks."""

    num_input_blocks = 1


    def propagate(
        self,
        from_blocks: List[Namespace],
        block: Namespace,
        propagators: dict,
        ext_vars: dict,
        cwd: str = None,
    ):
        """Method that propagates shapes in the given block.

        Args:
            from_blocks: The input blocks.
            block: The block to propagate its shapes.
            propagators: Dictionary of propagators.
            ext_vars: Dictionary of external variables required to load jsonnet.
            cwd: Working directory to resolve relative paths.

        Raises:
            ValueError: If there are multiple blocks with the same id.
            ValueError: If no propagator found for some block.
        """
        add_ids_prefix(block, from_blocks)
        blocks = get_blocks_dict(from_blocks + block.blocks)
        topological_predecessors = parse_graph(from_blocks, block)
        try:
            propagate_shapes(blocks,
                             topological_predecessors,
                             propagators=propagators,
                             ext_vars=ext_vars,
                             cwd=cwd)
        except Exception as ex:
            raise type(ex)(f'block[id={block._id}]: {ex}') from ex
        in_shape = get_shape('out', from_blocks[0])
        out_shape = get_shape('out', block.blocks[-1])
        block._shape = create_shape(in_shape, out_shape)


class GroupPropagator(SequentialPropagator):
    """Propagator for a sequence of blocks."""

    def propagate(
        self,
        from_blocks: List[Namespace],
        block: Namespace,
        propagators: dict,
        ext_vars: dict,
        cwd: str = None,
    ):
        """Method that propagates shapes in the given block.

        Args:
            from_blocks: The input blocks.
            block: The block to propagate its shapes.
            propagators: Dictionary of propagators.
            ext_vars: Dictionary of external variables required to load jsonnet.
            cwd: Working directory to resolve relative paths.

        Raises:
            ValueError: If there are multiple blocks with the same id.
            ValueError: If there graph references an undefined block.
            ValueError: If no propagator found for some block.
        """
        add_ids_prefix(block, from_blocks)
        blocks = get_blocks_dict(from_blocks + block.blocks)
        topological_predecessors = parse_graph(from_blocks, block)
        try:
            propagate_shapes(blocks,
                             topological_predecessors,
                             propagators=propagators,
                             ext_vars=ext_vars,
                             cwd=cwd)
        except Exception as ex:
            raise type(ex)(f'block[id={block._id}]: {ex}') from ex
        in_shape = get_shape('out', from_blocks[0])
        out_shape = get_shape('out', next(x for x in block.blocks if x._id==block.output))
        block._shape = create_shape(in_shape, out_shape)



"""Propagator classes for recurrent blocks."""

from jsonargparse import Namespace
from typing import List
from .base import BasePropagator, get_shape, create_shape, set_shape_dim
from ..schemas import auto_tag


class RnnPropagator(BasePropagator):
    """Propagator for recurrent style blocks."""

    num_input_blocks = 1
    output_feats_dims = 1


    def initial_checks(self, from_blocks: List[Namespace], block: Namespace):
        """Method that does some initial checks before propagation.

        Calls the base class checks and makes sure that the input shape has two
        dimensions and that block includes a valid output_feats attribute.

        Args:
            from_blocks: The input blocks.
            block: The block to propagate its shapes.

        Raises:
            ValueError: When block.output_feats not valid.
            ValueError: When len(from_block[0]._shape) != 2.
        """
        super().initial_checks(from_blocks, block)
        shape_in = get_shape('out', from_blocks[0])
        if len(shape_in) != 2:
            raise ValueError(f'{block._class} blocks require input shape to have 2 dimensions, but got {shape_in}.')


    def propagate(self, from_blocks: List[Namespace], block: Namespace):
        """Method that propagates shapes to a block.

        Args:
            from_blocks: The input blocks.
            block: The block to propagate its shapes.

        Raises:
            ValueError: When bidirectional==True and output_feats not even.
        """
        ## Set default values ##
        if not hasattr(block, 'bidirectional'):
            block.bidirectional = False

        ## Initialize block._shape ##
        from_shape = get_shape('out', from_blocks[0])
        output_feats = block.output_feats
        block._shape = create_shape(from_shape, [auto_tag, output_feats])

        ## Set hidden size ##
        if block.bidirectional and output_feats % 2 != 0:
            raise ValueError(f'For bidirectional {block._class} expected output_feats to be even, but got {output_feats}.')
        block.hidden_size = output_feats // (2 if block.bidirectional else 1)

        ## Propagate first dimension ##
        set_shape_dim('out', block, 0, from_shape[0])



"""Generic code for module architecture instantiators."""

import inspect
from ..schemas import mappings_validator, id_separator
from ..propagators.base import get_shape


def id_strip_parent_prefix(value):
    """Removes the parent prefix from an id value."""
    return value.rsplit(id_separator)[-1]


def import_object(name):
    """Function that returns a class in a module given its dot import statement."""
    name_module, name_class = name.rsplit('.', 1)
    module = __import__(name_module, fromlist=[name_class])
    return getattr(module, name_class)


def instantiate_block(block_cfg, blocks_mappings, module_cfg):
    """Function that instantiates a block given its narchi config and a mappings object."""
    mappings_validator.validate(blocks_mappings)
    if block_cfg._class not in blocks_mappings:
        raise NotImplementedError(f'No mapping for blocks of type {block_cfg._class}.')

    kwargs = {k: v for k, v in vars(block_cfg).items() if not k.startswith('_')}

    def set_kwargs(key_to, key_from, value=None):
        if key_to in kwargs:
            print(f'warning: mapping defines {key_to} as {key_from} so replacing current value: {kwargs[key_to]}.')
        if value is None:
            kwargs[key_to] = kwargs.pop(key_from)
        else:
            kwargs[key_to] = value

    block_mapping = blocks_mappings[block_cfg._class]
    block_class = import_object(block_mapping['class'])
    if 'kwargs' in block_mapping:
        for key_to, key_from in block_mapping['kwargs'].items():
            if key_to == ':skip:':
                del kwargs[key_from]
            elif key_from[0] == '_':
                set_kwargs(key_to, key_from, getattr(block_cfg, key_from))
            elif key_from.startswith('shape:'):
                _, key, idx = key_from.split(':')
                set_kwargs(key_to, key_from, get_shape(key, block_cfg)[int(idx)])
            elif key_from.startswith('const:'):
                _, vtype, val = key_from.split(':')
                if vtype == 'int':
                    val = int(val)
                elif vtype == 'bool':
                    val = False if val == 'False' else True
                set_kwargs(key_to, key_from, val)
            else:
                set_kwargs(key_to, key_from)

    if block_cfg._class == 'Module':
        set_kwargs('cfg', 'module_cfg', module_cfg)

    func_param = {x.name for x in inspect.signature(block_class).parameters.values()}
    if 'blocks_mappings' in func_param:
        set_kwargs('blocks_mappings', 'function_parameter', blocks_mappings)
    if 'module_cfg' in func_param:
        set_kwargs('module_cfg', 'function_parameter', module_cfg)

    return block_class(**kwargs)



"""Definition of the narchi json schemas."""

import json
import enum
from copy import deepcopy
from jsonschema import Draft7Validator as jsonvalidator


id_pattern = '[A-Za-z_][0-9A-Za-z_]*'
id_separator = '·'
propagated_id_pattern = '[A-Za-z_][0-9A-Za-z_'+id_separator+']*'
variable_pattern = '<<variable:([-+/*0-9A-Za-z_]+)>>'
auto_tag = '<<auto>>'


id_type = {
    'type': 'string',
    'pattern': '^'+id_pattern+'$',
}


description_type = {
    'type': 'string',
    'minLength': 8,
    'pattern': '^[^<>]+$',
}


dims_type = {
    'type': 'array',
    'minItems': 1,
    'items': {
        'oneOf': [
            {
                'type': 'integer',
                'minimum': 1,
            },
            {
                'type': 'string',
                'pattern': '^('+variable_pattern+'|'+auto_tag+')$',
            },
        ],
    },
}
dims_in_type = deepcopy(dims_type)
dims_in_type['items']['oneOf'].append({'type': 'null'})


shape_type = {
    'type': 'object',
    'properties': {
        'in':  {'$ref': '#/definitions/dims_in'},
        'out': {'$ref': '#/definitions/dims'},
    },
    'required': ['in', 'out'],
    'additionalProperties': False,
}


graph_type = {
    'type': 'array',
    'minItems': 1,
    'items': {
        'type': 'string',
        'pattern': '^'+id_pattern+'( +-> +'+id_pattern+')+$',
    },
}


path_type = {
    'type': 'string',
    'pattern': '.+\\.jsonnet',
}


reshape_dims_type = deepcopy(dims_type)
reshape_dims_type['minItems'] = 2
reshape_index_type = {
    'type': 'integer',
    'minimum': 0,
}
reshape_flatten_type = {
    'type': 'array',
    'minItems': 2,
    'items': {'$ref': '#/definitions/reshape_index'},
}
reshape_unflatten_type = {
    'type': 'object',
    'minProperties': 1,
    'maxProperties': 1,
    'patternProperties': {
        '^[0-9]+$': {'$ref': '#/definitions/reshape_dims'},
    },
    'additionalProperties': False,
}
reshape_type = {
    'oneOf': [
        {
            'const': 'flatten',
        },
        {
            'type': 'array',
            'minItems': 1,
            'items': {
                'oneOf': [
                    {'$ref': '#/definitions/reshape_index'},
                    {'$ref': '#/definitions/reshape_flatten'},
                    {'$ref': '#/definitions/reshape_unflatten'},
                ],
            },
        },
    ],
}
reshape_definitions = {
    'reshape': reshape_type,
    'reshape_index': reshape_index_type,
    'reshape_dims': reshape_dims_type,
    'reshape_flatten': reshape_flatten_type,
    'reshape_unflatten': reshape_unflatten_type,
}
reshape_schema = {
    '$ref': '#/definitions/reshape',
    'definitions': reshape_definitions,
}


block_type = {
    'type': 'object',
    'properties': {
        '_class':       {'$ref': '#/definitions/id'},
        '_name':        {'$ref': '#/definitions/id'},
        '_id':          {'$ref': '#/definitions/id'},
        '_id_share':    {'$ref': '#/definitions/id'},
        '_description': {'$ref': '#/definitions/description'},
        '_shape':       {'$ref': '#/definitions/shape'},
        '_path':        {'$ref': '#/definitions/path'},
        '_ext_vars':    {'type': 'object'},
        'blocks':       {'$ref': '#/definitions/blocks'},
        'input':        {'$ref': '#/definitions/id'},
        'output':       {'$ref': '#/definitions/id'},
        'graph':        {'$ref': '#/definitions/graph'},
        'dim':          {'type': 'integer'},
        'reshape_spec': {'$ref': '#/definitions/reshape'},
        'architecture': {'$ref': '#/definitions/architecture'},
    },
    #'required': ['_class', '_id'],
    'required': ['_class'],
    'allOf': [
        {
            'if': {'properties': {'_class': {'enum': ['Sequential', 'Group']}}},
            'then': {'required': ['blocks']},
            'else': {'not': {'required': ['blocks']}},
        },
        {
            'if': {'properties': {'_class': {'const': 'Group'}}},
            'then': {'required': ['graph', 'input', 'output']},
            'else': {'not': {'required': ['graph', 'input', 'output']}},
        },
        {
            'if': {'properties': {'_class': {'const': 'Module'}}},
            'then': {'required': ['_path']},
            'else': {'not': {'required': ['_path', '_ext_vars', 'architecture']}},
        },
        {
            'if': {'properties': {'_class': {'const': 'Sequential'}}},
            'else': {'properties': {'blocks': {'items': {'required': ['_id']}}}},  # not working!
        },
        {
            'if': {'properties': {'_class': {'const': 'Concatenate'}}},
            'then': {'required': ['dim']},
        },
        {
            'if': {'properties': {'_class': {'const': 'Reshape'}}},
            'then': {'required': ['reshape_spec']},
            'else': {'not': {'required': ['reshape_spec']}},
        },
    ],
}


blocks_type = {
    'type': 'array',
    'minItems': 1,
    'items': {'$ref': '#/definitions/block'},
}


inputs_outputs_type = {
    'type': 'array',
    'minItems': 1,
    'items': {
        'type': 'object',
        'properties': {
            '_id':          {'$ref': '#/definitions/id'},
            '_description': {'$ref': '#/definitions/description'},
            '_shape':       {'$ref': '#/definitions/dims'},
        },
        'required': ['_id', '_shape'],
        'additionalProperties': False,
    },
}


architecture_type = {
    'type': 'object',
    'properties': {
        '_id':          {'$ref': '#/definitions/id'},
        '_description': {'$ref': '#/definitions/description'},
        'blocks':       {'$ref': '#/definitions/blocks'},
        'graph':        {'$ref': '#/definitions/graph'},
        'inputs':       {'$ref': '#/definitions/inputs_outputs'},
        'outputs':      {'$ref': '#/definitions/inputs_outputs'},
    },
    'required': ['_id', 'blocks', 'graph', 'inputs', 'outputs'],
    'additionalProperties': False,
}


definitions = {
    'id': id_type,
    'description': description_type,
    'dims': dims_type,
    'dims_in': dims_in_type,
    'shape': shape_type,
    'graph': graph_type,
    'block': block_type,
    'blocks': blocks_type,
    'path': path_type,
    'inputs_outputs': inputs_outputs_type,
    'architecture': architecture_type,
}
definitions.update(reshape_definitions)


narchi_schema = {
    '$schema': 'http://json-schema.org/draft-07/schema#',
    '$id': 'https://schema.omnius.com/json/narchi/1.0/schema.json',
    'title': 'Neural Network Module Architecture Schema',
    '$ref': '#/definitions/architecture',
    'definitions': definitions,
}


block_definitions = deepcopy(definitions)
block_definitions['id'] = {'type': 'string', 'pattern': '^'+propagated_id_pattern+'$'}
block_definitions['graph']['items']['pattern'] = '^'+propagated_id_pattern+'( +-> +'+propagated_id_pattern+')+$'


block_schema = {
    '$ref': '#/definitions/block',
    'definitions': block_definitions,
}


propagated_definitions = deepcopy(block_definitions)
propagated_definitions['block']['required'] += ['_shape']
propagated_definitions['dims']['items']['oneOf'][1]['pattern'] = '^'+variable_pattern+'$'
propagated_definitions['architecture']['properties']['_shape'] = {'$ref': '#/definitions/shape'}
propagated_schema = deepcopy(narchi_schema)
propagated_schema['definitions'] = propagated_definitions
propagated_schema['title'] = 'Neural Network Module Propagated Architecture Schema'
del propagated_schema['$id']


mappings_schema = {
    'type': 'object',
    'minProperties': 1,
    'patternProperties': {
        '^'+id_pattern+'$': {
            'type': 'object',
            'properties': {
                'class': {
                    'type': 'string',
                    'pattern': '^[A-Za-z_][0-9A-Za-z_.]*$'
                },
                'kwargs': {
                    'type': 'object',
                    'patternProperties': {
                        '^'+id_pattern+'$': {
                            'type': 'string',
                            'oneOf': [
                                {'pattern': '^'+id_pattern+'$'},
                                {'pattern': '^shape:in:(|-)[0-9]+$'},
                                {'pattern': '^const:str:[^:]+$'},
                                {'pattern': '^const:int:[0-9]+$'},
                                {'pattern': '^const:bool:(True|False)$'},
                            ],
                        },
                        '^:skip:$': {
                            'type': 'string',
                            'pattern': '^'+id_pattern+'$',
                        },
                    },
                    'additionalProperties': False,
                },
                'required': ['class'],
                'additionalProperties': False,
            },
        },
    },
    'additionalProperties': False,
}


block_validator = jsonvalidator(block_schema)
narchi_validator = jsonvalidator(narchi_schema)
propagated_validator = jsonvalidator(propagated_schema)
reshape_validator = jsonvalidator(reshape_schema)
mappings_validator = jsonvalidator(mappings_schema)


class SchemasEnum(enum.Enum):
    """Enum of the schemas defined in narchi."""

    narchi = narchi_schema
    """Main schema which defines the general format for architecture files."""

    propagated = propagated_schema
    """Schema for architectures in which the dimensions have been propagated."""

    reshape = reshape_schema
    """Schema that defines the format to specify reshaping of tensors."""

    block = block_schema
    """Schema for a single architecture block."""

    mappings = mappings_schema
    """Schema for mappings between architectures and block implementations."""


schemas = {k: v.value for k, v in SchemasEnum.__dict__.items() if not k.startswith('_')}
schemas[None] = narchi_schema


def schema_as_str(schema: str = None):
    """Formats a schema as a pretty printed json string.

    Args:
        schema: The schema name to return among {'narchi', 'propagated', 'reshape', 'block', 'mappings'}.

    Returns:
        str: Pretty printed schema.
    """
    return json.dumps(schemas[schema], indent=2, ensure_ascii=False)



#!/usr/bin/env python3
"""General command line tool for narchi package functionalities."""

import sys
from jsonargparse import ArgumentParser, ActionPath
from narchi.render import ModuleArchitecture, ModuleArchitectureRenderer
from narchi.schemas import schema_as_str, schemas
from narchi import __version__


def get_parser():
    """Returns the argument parser object for the command line tool."""
    ## validate parser ##
    parser_validate = ModuleArchitecture.get_config_parser()
    parser_validate.description = 'Command for checking the validity of neural network module architecture files.'
    parser_validate.set_defaults(propagators='default')
    parser_validate.add_argument('jsonnet_paths',
        action=ActionPath(mode='fr'),
        nargs='+',
        help='Path(s) to neural network module architecture file(s) in jsonnet narchi format.')

    ## render parser ##
    parser_render = ModuleArchitectureRenderer.get_config_parser()
    parser_render.description = 'Command for rendering a neural network module architecture file.'
    parser_render.set_defaults(propagators='default')
    parser_render.add_argument('jsonnet_path',
        action=ActionPath(mode='fr'),
        help='Path to a neural network module architecture file in jsonnet narchi format.')
    parser_render.add_argument('out_file',
        nargs='?',
        action=ActionPath(mode='fc'),
        help='Path where to write the architecture diagram (with a valid extension for pygraphviz draw). If '
             'unset a pdf is saved to the output directory.')

    ## schema parser ##
    parser_schema = ArgumentParser(
        description='Prints a schema as a pretty json.')
    parser_schema.add_argument('schema',
        nargs='?',
        default='narchi',
        choices=[x for x in schemas.keys() if x is not None],
        help='Which of the available schemas to print.')

    ## global parser ##
    parser = ArgumentParser(
        description=__doc__,
        version=__version__)
    parser.add_argument('--stack_trace',
        type=bool,
        default=False,
        help='Whether to print stack trace when there are errors.')
    parser.parser_validate = parser_validate
    parser.parser_render = parser_render
    parser.parser_schema = parser_schema

    subcommands = parser.add_subcommands()
    subcommands.add_subcommand('validate', parser_validate)
    subcommands.add_subcommand('render', parser_render)
    subcommands.add_subcommand('schema', parser_schema)

    return parser


def get_validate_parser():
    return get_parser().parser_validate


def get_render_parser():
    return get_parser().parser_render


def get_schema_parser():
    return get_parser().parser_schema


def narchi_cli(argv=None):
    """Main execution function."""

    ## Parse arguments ##
    parser = get_parser()
    cfg = parser.parse_args(sys.argv[1:] if argv is None else argv)

    try:
        ## Schema subcommand ##
        if cfg.subcommand == 'schema':
            print(schema_as_str(cfg.schema.schema))

        ## Validate subcommand ##
        elif cfg.subcommand == 'validate':
            module = ModuleArchitecture(cfg=cfg.validate, parser=parser.parser_validate)
            for jsonnet_path in cfg.validate.jsonnet_paths:
                module.load_architecture(jsonnet_path)

        ## Render subcommand ##
        elif cfg.subcommand == 'render':
            if cfg.render.out_file is None:
                cfg.render.save_pdf = True
            module = ModuleArchitectureRenderer(cfg=cfg.render, parser=parser.parser_render)
            module.render(architecture=cfg.render.jsonnet_path, out_render=cfg.render.out_file)

    except Exception as ex:
        if cfg.stack_trace:
            raise ex
        print(f'error: {ex}')
        sys.exit(True)


## Main block called only when run from command line ##
if __name__ == '__main__':
    narchi_cli()



"""A standard module instantiator for pytorch."""

import torch
from functools import reduce
from collections import OrderedDict
from jsonargparse import Path, get_config_read_mode
from typing import List

from .common import instantiate_block, id_strip_parent_prefix
from ..module import ModuleArchitecture
from ..propagators.reshape import check_reshape_spec, norm_reshape_spec
from ..propagators.group import get_blocks_dict
from ..graph import parse_graph
from ..schemas import auto_tag


class BaseModule(ModuleArchitecture, torch.nn.Module):
    """Base class for instantiation of pytorch modules based on a narchi architecture."""

    def __init__(
        self,
        *args,
        state_dict: dict = None,
        debug: bool = False,
        **kwargs
    ):
        """Initializer for BaseModule class.

        Args:
            state_dict: State dictionary to set.
            debug: Enable to keep self.intermediate_outputs.
            args/kwargs: All other arguments accepted by :class:`.ModuleArchitecture`.
        """
        torch.nn.Module.__init__(self)
        ModuleArchitecture.__init__(self, *args, **kwargs)

        architecture = self.architecture
        blocks = self.blocks
        module_cfg = {'propagated': True, 'cwd': self.cfg.cwd}
        if 'cfg' in kwargs and 'ext_vars' in kwargs['cfg']:
            module_cfg['ext_vars'] = kwargs['cfg']['ext_vars']

        io_ids = {x._id for x in architecture.inputs+architecture.outputs}

        for block_id in blocks.keys():
            if block_id not in io_ids:
                block = instantiate_block(blocks[block_id], self.blocks_mappings, module_cfg)
                setattr(self, id_strip_parent_prefix(block_id), block)

        self.state_dict_prop = state_dict
        self.debug = debug


    def forward(self, *args, **kwargs):
        """Runs a forward using the architecture's inputs and graph."""
        inputs = {x for x in kwargs.keys()}
        expected_inputs = {x._id for x in self.architecture.inputs}
        if len(args) != 0:
            raise RuntimeError(f'{type(self).__name__} expects only keyword arguments.')
        if inputs != expected_inputs:
            raise RuntimeError(f'{type(self).__name__} got unexpected arguments, given {inputs}, expected {expected_inputs}.')

        values = OrderedDict({x: kwargs[x] for x in inputs})
        self.inputs_preprocess(values)
        self.check_inputs_shape(values)
        device = next(self.parameters()).device
        for key, value in values.items():
            values[key] = value.to(device)

        out_ids = {x._id for x in self.architecture.outputs}
        graph_forward(self, values, out_ids, intermediate_outputs=self.debug)

        if len(self.architecture.outputs) == 1:
            return values[self.architecture.outputs[0]._id]
        else:
            return [values[x._id] for x in self.architecture.outputs]


    def inputs_preprocess(self, values):
        """Pre-processing for inputs, base implementation does nothing."""
        pass


    def check_inputs_shape(self, values):
        """Raises error if the shape of any input does not agree with architecture."""
        for node in self.architecture.inputs:
            given = self.get_tensor_shape(values[node._id])
            expected = node._shape
            if len(expected) != len(given) \
               or any(isinstance(expected[n], int) and given[n] != expected[n] for n in range(len(given))):
                raise RuntimeError(f'{type(self).__name__} got unexpected tensor shape for input "{node._id}", '
                                   f'given {given}, expected {expected}.')


    def get_tensor_shape(self, value) -> List[int]:
        """Returns tensor shape excluding batch dimension."""
        return list(value.shape[1:])


    @property
    def state_dict_prop(self):
        """The current state dictionary."""
        return super().state_dict()


    @state_dict_prop.setter
    def state_dict_prop(self, state_dict: dict):
        """Replaces the current state dictionary with the one given.

        Args:
            state_dict: State dictionary to set.
        """
        if state_dict is None:
            return
        elif isinstance(state_dict, (str, Path)):
            path = Path(state_dict, mode=get_config_read_mode(), cwd=self.cfg.cwd)
            state_dict = torch.load(path())
        if not isinstance(state_dict, dict):
            raise ValueError('Expected state_dict to be a dictionary.')

        self.load_state_dict(state_dict)


class Sequential(torch.nn.Sequential):
    """Sequential module that receives as input an narchi blocks object."""
    def __init__(self, blocks, blocks_mappings, module_cfg):
        subblock_list = []
        for subblock in blocks:
            subblock_list.append(instantiate_block(subblock, blocks_mappings, module_cfg))
        super().__init__(*subblock_list)


class Add(torch.nn.Module):
    """Module that adds all inputs element-wise, thus expects all inputs to be the same shape."""
    def forward(self, *inputs):
        return sum(inputs)


class Concatenate(torch.nn.Module):
    """Module that concatenates input tensors along a given dimension."""
    def __init__(self, dim):
        super().__init__()
        self.dim = dim if dim < 0 else dim+1

    def forward(self, *inputs):
        return torch.cat(inputs, self.dim)  # pylint: disable=no-member


class LSTM(torch.nn.LSTM):
    """Extension of torch.nn.LSTM that allows disabling the return of the hidden and cell states."""
    def __init__(self, *args, output_state=False, **kwargs):
        self.output_state = output_state
        super().__init__(*args, **kwargs)

    def forward(self, input):
        output = super().forward(input)
        return output if self.output_state else output[0]


class Reshape(torch.nn.Module):
    """Reshape module configurable by a narchi reshape_spec object."""

    def __init__(self, reshape_spec):
        super().__init__()
        self.reshape_spec = norm_reshape_spec(reshape_spec)
        self.idxs = check_reshape_spec(self.reshape_spec)

    def forward(self, input):
        """Transforms the shape of the input according to the specification in reshape_spec."""
        in_dims = input.shape[1:]
        idxs = self.idxs
        if self.reshape_spec == 'flatten':
            idxs = [n for n in range(len(input.shape)-1)]
        if len(input.shape) != len(idxs)+1:
            raise RuntimeError(f'{type(self).__name__} got a tensor with {len(input.shape)} dimensions but expected {len(idxs)+1}.')

        reshaped = input
        if idxs != [x for x in range(len(idxs))]:
            permute = [0] + [x+1 for x in idxs]
            reshaped = reshaped.permute(*permute)

        def prod(values):
            return reduce((lambda x, y: x * y), values)

        reshape_spec = self.reshape_spec
        if self.reshape_spec == 'flatten':
            reshape_spec = [idxs]
        if any(isinstance(x, (list, dict)) for x in reshape_spec):
            reshape = [input.shape[0]]
            for val in reshape_spec:
                if isinstance(val, int):
                    reshape.append(in_dims[val])
                elif isinstance(val, list):
                    reshape.append(prod(in_dims[v] for v in val))
                elif isinstance(val, dict):
                    idx = next(iter(val.keys()))
                    in_dim = in_dims[int(idx)]
                    dims = val[idx]
                    if any(x == auto_tag for x in dims):
                        auto_idx = dims.index(auto_tag)
                        nonauto = prod(x for x in dims if x != auto_tag)
                        dims[auto_idx] = in_dim//nonauto
                    reshape.extend(dims)

            reshaped = reshaped.reshape(reshape)

        return reshaped


class Group(torch.nn.Module):
    """Group module that receives narchi blocks, graph, input and output objects."""

    def __init__(self, block_id, blocks, blocks_mappings, module_cfg, graph, input, output):
        super().__init__()

        from_input = block_id+'¦input'
        from_blocks = [{'_id': from_input}]
        block = {'_id': block_id, 'blocks': blocks, 'graph': graph, 'input': input}
        self.topological_predecessors = parse_graph(from_blocks, block)
        self.input = from_input
        self.output = output
        self.blocks_mappings = blocks_mappings
        self.blocks = get_blocks_dict(blocks)

        for num in range(len(blocks)):
            block_id = id_strip_parent_prefix(blocks[num]._id)
            block = instantiate_block(blocks[num], blocks_mappings, module_cfg)
            setattr(self, block_id, block)


    def forward(self, input):
        """Runs a forward using the group's input and graph."""
        device = next(self.parameters()).device
        values = OrderedDict({self.input: input.to(device)})
        graph_forward(self, values)
        return values[self.output]


def graph_forward(module, values, out_ids=set(), intermediate_outputs=False):
    """Runs a forward for a module using its graph topological_predecessors."""

    if intermediate_outputs and not hasattr(module, 'intermediate_outputs'):
        module.intermediate_outputs = OrderedDict()

    needed_until = OrderedDict()
    for node, inputs in module.topological_predecessors.items():
        for input in inputs:
            needed_until[input] = node

    for node, inputs in module.topological_predecessors.items():
        if node in out_ids:
            values[node] = values[inputs[0]]
            continue

        submodule = getattr(module, id_strip_parent_prefix(node))
        try:
            if isinstance(submodule, BaseModule):
                assert len(inputs) == 1
                assert len(submodule.architecture.inputs) == 1
                result = submodule(**{submodule.architecture.inputs[0]._id: next(iter(values.values()))})
            else:
                result = submodule(*[values[x] for x in inputs])
        except Exception as ex:
            raise type(ex)(f'{type(submodule).__name__}[id={node}]: {ex}') from ex
        values[node] = result

        if intermediate_outputs:
            module.intermediate_outputs[node] = result

        for input in inputs:
            if needed_until[input] == node:
                del values[input]


standard_pytorch_blocks_mappings = {
    'Sequential': {
        'class': 'narchi.instantiators.pytorch.Sequential',
    },
    'Add': {
        'class': 'narchi.instantiators.pytorch.Add',
    },
    'Concatenate': {
        'class': 'narchi.instantiators.pytorch.Concatenate',
    },
    'LSTM': {
        'class': 'narchi.instantiators.pytorch.LSTM',
        'kwargs': {
            'input_size': 'shape:in:1',
            'batch_first': 'const:bool:True',
            ':skip:': 'output_feats',
        },
    },
    'Reshape': {
        'class': 'narchi.instantiators.pytorch.Reshape',
    },
    'Group': {
        'class': 'narchi.instantiators.pytorch.Group',
        'kwargs': {
            'block_id': '_id',
        },
    },
    'Module': {
        'class': 'narchi.instantiators.pytorch.StandardModule',
    },
    'Softmax': {
        'class': 'torch.nn.Softmax',
    },
    'LogSoftmax': {
        'class': 'torch.nn.LogSoftmax',
    },
    'ReLU': {
        'class': 'torch.nn.ReLU',
    },
    'LeakyReLU': {
        'class': 'torch.nn.LeakyReLU',
    },
    'Dropout': {
        'class': 'torch.nn.Dropout',
    },
    'Conv1d': {
        'class': 'torch.nn.Conv1d',
        'kwargs': {
            'in_channels': 'shape:in:0',
            'out_channels': 'output_feats',
        },
    },
    'Conv2d': {
        'class': 'torch.nn.Conv2d',
        'kwargs': {
            'in_channels': 'shape:in:0',
            'out_channels': 'output_feats',
        },
    },
    'BatchNorm2d': {
        'class': 'torch.nn.BatchNorm2d',
        'kwargs': {
            'num_features': 'shape:in:0',
        },
    },
    'MaxPool1d': {
        'class': 'torch.nn.MaxPool1d',
    },
    'MaxPool2d': {
        'class': 'torch.nn.MaxPool2d',
        #'kwargs': {
        #    'ceil_mode': 'const:bool:True',
        #},
    },
    'AdaptiveAvgPool1d': {
        'class': 'torch.nn.AdaptiveAvgPool1d',
        'kwargs': {
            'output_size': 'output_feats',
        },
    },
    'AdaptiveAvgPool2d': {
        'class': 'torch.nn.AdaptiveAvgPool2d',
        'kwargs': {
            'output_size': 'output_feats',
        },
    },
    'Embedding': {
        'class': 'torch.nn.Embedding',
        'kwargs': {
            'embedding_dim': 'output_feats',
        },
    },
    'Identity': {
        'class': 'torch.nn.Identity',
    },
    'Linear': {
        'class': 'torch.nn.Linear',
        'kwargs': {
            'in_features': 'shape:in:-1',
            'out_features': 'output_feats',
        },
    },
}


class StandardModule(BaseModule):
    """Class for instantiating pytorch modules with standard block mappings."""
    blocks_mappings = standard_pytorch_blocks_mappings



"""Propagator classes for convolution blocks."""

from jsonargparse import Namespace
from typing import List
from .base import BasePropagator, get_shape, create_shape, set_shape_dim, check_output_feats_dims
from ..sympy import conv_out_length
from ..schemas import auto_tag


class ConvPropagator(BasePropagator):
    """Propagator for convolution style blocks."""

    num_input_blocks = 1
    num_features_source = 'output_feats'
    conv_dims = None


    def __init__(self, block_class: str, conv_dims: int):
        """Initializer for ConvPropagator instance.

        Args:
            block_class: The name of the block class being propagated.
            conv_dims: Number of dimensions for the convolution.

        Raises:
            ValueError: If conv_dims not int > 0.
        """
        super().__init__(block_class)
        valid_num_features_source = {'output_feats', 'from_shape'}
        if self.num_features_source not in valid_num_features_source:
            raise ValueError(f'{type(self).__name__} only allows num_features_source to be one of {valid_num_features_source}.')
        if not isinstance(conv_dims, int) or conv_dims < 1:
            raise ValueError(f'{type(self).__name__} only allows conv_dims to be an int > 0.')
        self.conv_dims = conv_dims


    def initial_checks(self, from_blocks: List[Namespace], block: Namespace):
        """Method that does some initial checks before propagation.

        Calls the base class checks and makes sure that the input shape agrees
        with the convolution dimensions.

        Args:
            from_blocks: The input blocks.
            block: The block to propagate its shapes.

        Raises:
            ValueError: When conv_dims does not agree with from_block[0]._shape.
        """
        super().initial_checks(from_blocks, block)
        shape_in = get_shape('out', from_blocks[0])
        if len(shape_in)-1 != self.conv_dims:
            raise ValueError(f'{block._class} blocks require input shape to have {self.conv_dims+1} dimensions, '
                             f'but got {shape_in}.')


    def propagate(self, from_blocks: List[Namespace], block: Namespace):
        """Method that propagates shapes to a block.

        Args:
            from_blocks: The input blocks.
            block: The block to propagate its shapes.

        Raises:
            ValueError: When block.output_feats not valid.
            NotImplementedError: If num_features_source is not one of {"from_shape", "output_feats"}.
        """
        ## Set default values ##
        kernel = block.kernel_size
        stride = block.stride if hasattr(block, 'stride') else 1
        padding = block.padding if hasattr(block, 'padding') else 0
        dilation = block.dilation if hasattr(block, 'dilation') else 1

        ## Initialize block._shape ##
        auto_dims = [auto_tag for _ in range(self.conv_dims)]
        from_shape = get_shape('out', from_blocks[0])
        if self.num_features_source == 'from_shape':
            block._shape = create_shape(from_shape, [from_shape[0]]+auto_dims)
        elif self.num_features_source == 'output_feats':
            check_output_feats_dims(1, self.block_class, block)
            block._shape = create_shape(from_shape, [block.output_feats]+auto_dims)

        ## Calculate and set <<auto>> output dimensions ##
        for dim, val in enumerate(get_shape('out', block)):
            if val == auto_tag:
                in_length = get_shape('in', block)[dim]
                out_length = conv_out_length(in_length, kernel, stride, padding, dilation)
                set_shape_dim('out', block, dim, out_length)


class PoolPropagator(ConvPropagator):
    """Propagator for pooling style blocks."""

    num_features_source = 'from_shape'



"""Propagator classes for concatenating."""

from jsonargparse import Namespace
from typing import List
from .base import BasePropagator, get_shape, create_shape
from ..sympy import sum


class ConcatenatePropagator(BasePropagator):
    """Propagator for concatenating along a given dimension."""

    num_input_blocks = '>1'


    def initial_checks(self, from_blocks: List[Namespace], block: Namespace):
        """Method that does some initial checks before propagation.

        Calls the base class checks and makes sure that the dim attribute
        is valid and agrees with the input dimensions.

        Args:
            from_blocks: The input blocks.
            block: The block to propagate its shapes.

        Raises:
            ValueError: When block does not have a valid dim attribute that agrees with input dimensions.
        """
        super().initial_checks(from_blocks, block)
        shape_0 = get_shape('out', from_blocks[0])
        dim = block.dim if block.dim >= 0 else len(shape_0)+block.dim
        if dim < 0 or dim >= len(shape_0):
            raise ValueError(f'Value of dim attribute ({block.dim}) in block[id={block._id}] does not '
                             f'agree with the input dimensions coming from block[id={from_blocks[0]._id}].')
        for n in range(1, len(from_blocks)):
            shape_n = get_shape('out', from_blocks[n])
            if len(shape_0) != len(shape_n) or \
               any(shape_0[k] != shape_n[k] for k in range(len(shape_0)) if k != dim):
                raise ValueError(f'{self.block_class} expects all inputs to have the same shape except along '
                                 f'the concatenating dimension, differs for block[id={from_blocks[n]._id}] '
                                 f'connecting to block[id={block._id}], {shape_0} vs. {shape_n}.')


    def propagate(self, from_blocks: List[Namespace], block: Namespace):
        """Method that propagates shapes to a block.

        Args:
            from_blocks: The input blocks.
            block: The block to propagate its shapes.
        """
        shape_in = list(get_shape('out', from_blocks[0]))
        shape_in[block.dim] = None
        shape_out = list(shape_in)
        shape_out[block.dim] = sum([get_shape('out', b)[block.dim] for b in from_blocks])
        block._shape = create_shape(shape_in, shape_out)



"""A pytorch module instantiator that supports 1d and 2d packed sequences."""

import math
import torch
import numpy as np
from copy import deepcopy
from collections import namedtuple
from torch.nn.utils.rnn import PackedSequence
from typing import List, Tuple, Union

from .pytorch import BaseModule, Reshape, standard_pytorch_blocks_mappings


class Packed2dSequence(namedtuple('Packed2dSequence', 'data lengths gaps')):
    """Named tuple for packed 2d sequences."""

    def to(self, *args, **kwargs):
        """Performs dtype and/or device conversion on `self.data`."""
        data = self.data.to(*args, **kwargs)
        return type(self)(data=data, lengths=self.lengths, gaps=self.gaps)


def pack_2d_sequences(
    input: Union[Tuple[torch.Tensor], List[torch.Tensor]],
    gap_size: int = 0,
    length_fact: int = 1,
    fail_if_unsorted: bool = True,
) -> Packed2dSequence:
    """Packs 3-dim tensors into a long 3-dim tensor by concatenating along the last dimension.

    Args:
        input: Input tensors to pack with shapes [C, H, Wi].
        gap_size: Size for gap between samples.
        length_fact: Increase gaps so that length of inputs are multiple of length_fact.
        fail_if_unsorted: Whether to raise ValueError if received unsorted input.

    Returns:
        Packed2dSequence(data=tensor[1, C, H, lengths+gaps], lenghts=list[len(input)], gaps=[len(input)])
    """
    if not (isinstance(input, (tuple, list))
            and all([hasattr(x, 'shape') for x in input])
            and all([x.dim() == 3 for x in input])
            and all([x.shape[0] == input[0].shape[0] for x in input])
            and all([x.shape[1] == input[0].shape[1] for x in input])):
        raise RuntimeError('Input required to be a tuple/list of tensors all with 3 dimensions and '
                           'have the same size for the first two dimensions.')

    ## Check that input is sorted from longest to shortest ##
    lengths = np.array([x.shape[2] for x in input], dtype=int)
    if any([d > 0 for d in np.diff(lengths)]):
        if fail_if_unsorted:
            raise ValueError('Expected input to be sorted from longest to shortest.')
        input = list(input)
        input.sort(key=lambda x: x.shape[2], reverse=True)
        lengths = np.array([x.shape[2] for x in input], dtype=int)

    ## Determine gaps ##
    gaps = np.zeros(len(input), dtype=int)
    for num in range(len(input)):
        gaps[num] = gap_size if num < len(input)-1 else 0
        fact_off = (lengths[num]+gaps[num]) % length_fact
        if fact_off > 0:
            gaps[num] += length_fact - fact_off

    ## Create tensor for packing ##
    total_length = lengths.sum() + gaps.sum()
    packed = torch.zeros((1, *input[0].shape[0:2], total_length), dtype=input[0].dtype, device=input[0].device)  # pylint: disable=no-member

    ## Copy input to packed tensor ##
    offset = 0
    for num in range(len(input)):
        packed[0, :, :, offset:offset+lengths[num]] = input[num]
        offset += lengths[num] + gaps[num]

    return Packed2dSequence(data=packed, lengths=lengths, gaps=gaps)


def packed_2d_set_gaps_to_zero(packed: Packed2dSequence) -> Packed2dSequence:
    """Sets to zero the gap locations of a Packed2dSequence."""
    assert isinstance(packed, Packed2dSequence), 'Expected input to be a Packed2dSequence.'
    lengths = packed.lengths
    gaps = packed.gaps
    if len(lengths) > 1 and gaps.sum() > 0:
        offset = 0
        for num in range(len(lengths)):
            packed.data[0, :, :, offset+lengths[num]:offset+lengths[num]+gaps[num]] = 0.0
            offset += lengths[num] + gaps[num]
    return packed


def packed_2d_to_1d(packed_2d: Packed2dSequence) -> PackedSequence:
    """Converts a Packed2dSequence to a PackedSequence."""
    assert isinstance(packed_2d, Packed2dSequence), 'Expected input to be a Packed2dSequence.'

    data = packed_2d.data
    lengths = packed_2d.lengths
    gaps = packed_2d.gaps

    ## Reserve memory for packed 1d tensor ##
    packed_1d_length = lengths.sum()
    packed_1d_size = (packed_1d_length, np.prod(data.shape[1:3]))
    data_1d = torch.zeros(packed_1d_size, dtype=data.dtype, device=data.device)  # pylint: disable=no-member

    ## Get packed 2d sequence offsets and batch_sizes ##
    offsets = [0]
    batch_sizes = torch.zeros(lengths[0], dtype=torch.long)  # pylint: disable=no-member
    for num, length in enumerate(lengths):
        offsets.append(offsets[num]+length+gaps[num])
        batch_sizes[0:length] += 1

    ## Copy data to packed 1d tensor ##
    t = 0
    p = 0
    for nsamp in batch_sizes:
        for num in range(nsamp):
            data_1d[p, :] = data[0, :, :, offsets[num]+t].reshape(-1)
            p += 1
        t += 1

    ## Return PackedSequence ##
    sorted_indices = torch.tensor([num for num in range(len(lengths))], dtype=torch.long, device=data.device)  # pylint: disable=not-callable,no-member
    return PackedSequence(data_1d, batch_sizes, sorted_indices, sorted_indices)


class ReshapePacked(Reshape):
    """Reshape module that works with Packed2dSequence."""

    def forward(self, input):
        """Transforms the shape of the input according to the specification in reshape_spec."""
        if not isinstance(input, Packed2dSequence):
            return super().forward(input)
        if self.reshape_spec != [2, [0, 1]]:
            raise RuntimeError(f'Reshape of Packed2dSequence only supported for reshape_spec=[2, [0, 1]].')
        return packed_2d_to_1d(input)


class Conv2dPacked(torch.nn.Conv2d):
    """Extension of torch.nn.Conv2d that works with Packed2dSequence.

    After convolving, the gaps are set to zero to guaranty that these values are
    not considered when computing gradients or subsequent forwards.
    """

    def __init__(self, in_channels, out_channels, kernel_size, padding, bias=True):
        if kernel_size//2 != padding:
            raise NotImplementedError('Currently only implemented for kernel_size//2 == padding.')
        super().__init__(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            #stride=stride,
            padding=padding,
            #dilation=dilation,
            #groups=groups,
            bias=bias,
            padding_mode='zeros')
        self.min_gap = kernel_size//2

    def forward(self, input):
        if not isinstance(input, Packed2dSequence):
            return super().forward(input)
        if any(g<self.min_gap for g in input.gaps[:-1]):
            raise RuntimeError(f'Gaps too small to prevent interference between samples, min_gap={self.min_gap} '
                               f'gaps={list(input.gaps[:-1])}.')
        output_data = super().forward(input.data)
        #_save_image_detached(output_data[0,0:3,:,:], 'post_conv.png')
        output = packed_2d_set_gaps_to_zero(Packed2dSequence(data=output_data, lengths=input.lengths, gaps=input.gaps))
        #_save_image_detached(output.data[0,0:3,:,:], 'post_conv_mask.png')
        return output


class MaxPool2dPacked(torch.nn.MaxPool2d):
    """Extension of torch.nn.MaxPool2d that works with Packed2dSequence.

    The lengths and gaps are re-estimated.
    """

    def __init__(self, kernel_size, stride):
        if kernel_size != stride:
            raise NotImplementedError('Currently only implemented for kernel_size == stride.')
        super().__init__(kernel_size=kernel_size, stride=stride)

    def forward(self, input):
        if not isinstance(input, Packed2dSequence):
            return super().forward(input)

        ## Compute new packed lengths and gaps ##
        lengths = np.zeros(input.lengths.shape, dtype=int)
        gaps = np.zeros(input.gaps.shape, dtype=int)
        for num in range(len(lengths)):
            length_in = input.lengths[num]
            gap_in = input.gaps[num]
            if (length_in+gap_in) % self.kernel_size > 0:
                raise RuntimeError(f'Expected sum of lengths of sample and gap be a multiple of pooling kernel '
                                   f'size: num={num} kernel_size={self.kernel_size} length={length_in} gap={gap_in}.')
            #if self.ceil_mode:
            lengths[num] = math.ceil(length_in/self.kernel_size)
            #else:
            #    lengths[num] = math.floor(length_in/self.kernel_size)
            gaps[num] = math.floor(gap_in/self.kernel_size)

        ## Perform pooling ##
        #_save_image_detached(input.data[0,0:3,:,:], 'pre_maxpool.png')
        output_data = super().forward(input.data)
        #_save_image_detached(output_data[0,0:3,:,:], 'post_maxpool.png')

        ## Check that pooled tensor matches shape ##
        if lengths.sum()+gaps.sum() != output_data.shape[-1]:
            raise RuntimeError('Bug in implementation, pooling length differs w.r.t. lengths and gaps computation.')

        return Packed2dSequence(data=output_data, lengths=lengths, gaps=gaps)


class BatchNorm2dPacked(torch.nn.BatchNorm2d):
    """Extension of torch.nn.BatchNorm2d that works with Packed2dSequence.

    After normalization, the gaps are set to zero to guaranty that these values
    are not considered when computing gradients or subsequent forwards.
    """
    def forward(self, input):
        if not isinstance(input, Packed2dSequence):
            return super().forward(input)
        output_data = super().forward(input.data)
        #_save_image_detached(output_data[0,0:3,:,:], 'post_bnorm.png')
        output = packed_2d_set_gaps_to_zero(Packed2dSequence(data=output_data, lengths=input.lengths, gaps=input.gaps))
        #_save_image_detached(output.data[0,0:3,:,:], 'post_bnorm_mask.png')
        return output


class LeakyReLU2dPacked(torch.nn.LeakyReLU):
    """Extension of torch.nn.LeakyReLU that works with Packed2dSequence."""
    def forward(self, input):
        if not isinstance(input, Packed2dSequence):
            return super().forward(input)
        output_data = super().forward(input.data)
        return Packed2dSequence(data=output_data, lengths=input.lengths, gaps=input.gaps)


class Linear1dPacked(torch.nn.Linear):
    """Extension of torch.nn.Linear that works with PackedSequence."""
    def forward(self, input):
        if not isinstance(input, PackedSequence):
            return super().forward(input)
        output_data = super().forward(input.data)
        return PackedSequence(data=output_data, batch_sizes=input.batch_sizes, sorted_indices=input.sorted_indices)


class LogSoftmaxPacked(torch.nn.LogSoftmax):
    """Extension of torch.nn.LogSoftmax that works with PackedSequence."""
    def forward(self, input):
        if not isinstance(input, PackedSequence):
            return super().forward(input)
        output_data = super().forward(input.data)
        return PackedSequence(data=output_data, batch_sizes=input.batch_sizes, sorted_indices=input.sorted_indices)


mappings = {
    'Reshape': {
        'class': 'narchi.instantiators.pytorch_packed.ReshapePacked',
    },
    'Conv2d': {
        'class': 'narchi.instantiators.pytorch_packed.Conv2dPacked',
        'kwargs': {
            'in_channels': 'shape:in:0',
            'out_channels': 'output_feats',
        },
    },
    'MaxPool2d': {
        'class': 'narchi.instantiators.pytorch_packed.MaxPool2dPacked',
    },
    'BatchNorm2d': {
        'class': 'narchi.instantiators.pytorch_packed.BatchNorm2dPacked',
        'kwargs': {
            'num_features': 'shape:in:0',
        },
    },
    'LogSoftmax': {
        'class': 'narchi.instantiators.pytorch_packed.LogSoftmaxPacked',
    },
    'LeakyReLU': {
        'class': 'narchi.instantiators.pytorch_packed.LeakyReLU2dPacked',
        'kwargs': {
            'inplace': 'const:bool:True',
        },
    },
    'Linear': {
        'class': 'narchi.instantiators.pytorch_packed.Linear1dPacked',
        'kwargs': {
            'in_features': 'shape:in:-1',
            'out_features': 'output_feats',
        },
    },
}

packed_pytorch_blocks_mappings = deepcopy(standard_pytorch_blocks_mappings)
packed_pytorch_blocks_mappings.update(mappings)


class PackedModule(BaseModule):
    """Class for instantiating pytorch modules that support 1d and 2d packed sequences."""

    blocks_mappings = packed_pytorch_blocks_mappings

    def __init__(
        self,
        *args,
        gap_size: int = 0,
        length_fact: int = 1,
        **kwargs
    ):
        """Initializer for PackedModule class.

        Args:
            gap_size: Size for gap between samples.
            length_fact: Increase gaps so that length of inputs are multiple of length_fact.
            args/kwargs: All other arguments accepted by :class:`.BaseModule`.
        """
        BaseModule.__init__(self, *args, **kwargs)
        self.gap_size = gap_size
        self.length_fact = length_fact


    def inputs_preprocess(self, values: dict):
        """Converts tuples of tensors into Packed2dSequence.

        Args:
            values: Inputs to the module.
        """
        for key, value in values.items():
            if isinstance(value, (tuple, list)) and not isinstance(value, Packed2dSequence):
                values[key] = pack_2d_sequences(value, gap_size=self.gap_size, length_fact=self.length_fact)


    def get_tensor_shape(self, value) -> List[int]:
        if isinstance(value, Packed2dSequence):
            return list(value.data.shape[1:])
        return super().get_tensor_shape(value)



"""Functions for symbolic operations."""

import re
import sympy
from sympy.core import numbers
from typing import List, Union
from .schemas import variable_pattern


variable_regex = re.compile('^'+variable_pattern+'$')


def is_valid_dim(value):
    """Checks whether value is an int > 0 or str that follows variable_regex.pattern."""
    if isinstance(value, int) or variable_regex.match(str(value)):
        if isinstance(value, int) and value <= 0:
            return False
        return True
    return False


def sympify_variable(value):
    """Returns the sympyfied object for the given value."""
    var_match = variable_regex.match(str(value))
    if var_match:
        value_sympy = sympy.sympify(var_match[1])
    elif isinstance(value, (int, str)):
        value_sympy = sympy.sympify(value)
    else:
        raise ValueError(f'Expected input to be an int or a valid sympy expression or a string with '
                         f'pattern: {variable_regex.pattern}, but got {value}.')
    return value_sympy


def get_nonrational_variable(value):
    """Returns either an int or a string variable."""
    if isinstance(value, numbers.Integer):
        return int(value)
    elif isinstance(value, numbers.Rational):
        raise ValueError(f'Obtained a rational result: {value}.')
    return '<<variable:'+str(value).replace(' ', '')+'>>'


def variable_operate(value: Union[str, int], operation: Union[str, int]) -> Union[str, int]:
    """Performs a symbolic operation on a given value.

    Args:
        value: The value to operate on, either an int or a variable, e.g. "<<variable:W/2+H/4>>".
        operation: Operation to apply on value, either int or expression, e.g. "__input__/3".

    Returns:
        The result of the operation.

    Raises:
        ValueError:
            * If operation is not int nor a valid expression.
            * If value is not an int or a string that follows variable_regex.pattern.
            * If value is not a valid expression or contains "__input__" as a free symbol.
    """
    ## Handle operation ##
    if isinstance(operation, int):
        return operation
    elif not isinstance(operation, str):
        raise ValueError('Expected operation to be an int or a string.')
    operation_sympy = sympify_variable(operation)
    ## Handle input ##
    value_sympy = sympify_variable(value)
    value_symbs = {str(s) for s in value_sympy.free_symbols}
    if '__input__' in value_symbs:
        raise ValueError(f'Value must not contain "__input__" as a free symbol, but got {input}.')
    ## Substitute input into operation ##
    output_sympy = operation_sympy.subs({'__input__': value_sympy})
    ## Return operation result ##
    return get_nonrational_variable(output_sympy)


def variables_aggregate(values: List[Union[str, int]], operation: str) -> Union[str, int]:
    """Performs a symbolic aggregation operation over all input values.

    Args:
        values: List of values to operate on.
        operation: One of '+'=sum, '*'=prod.

    Returns:
        The result of the operation.

    Raises:
        ValueError: If any value is not an int or a string that follows variable_regex.pattern.
    """
    operations = {'+', '*'}
    if operation not in operations:
        raise ValueError(f'Expected operation to be one of {operations}, got {operation}.')
    if not isinstance(values, list) or len(values) < 1 or not all(isinstance(v, (str, int)) for v in values):
        raise ValueError(f'Expected values to be a list containing int or str elements, got {values}.')
    value = values[0]
    for num in range(1, len(values)):
        value_sympy = sympify_variable(value)
        value = variable_operate(values[num], f'__input__{operation}({value_sympy})')
    return value


def sum(values: List[Union[str, int]]) -> Union[str, int]:
    """Performs a symbolic sum of all input values.

    Args:
        values: List of values to operate on.

    Returns:
        The result of the operation.

    Raises:
        ValueError: If any value is not an int or a string that follows variable_regex.pattern.
    """
    return variables_aggregate(values, '+')


def prod(values: List[Union[str, int]]) -> Union[str, int]:
    """Performs a symbolic product of all input values.

    Args:
        values: List of values to operate on.

    Returns:
        The result of the operation.

    Raises:
        ValueError: If any value is not an int or a string that follows variable_regex.pattern.
    """
    return variables_aggregate(values, '*')


def divide(numerator: Union[str, int], denominator: Union[str, int]) -> Union[str, int]:
    """Performs a symbolic division.

    Args:
        numerator: Value for numerator.
        denominator: Value for denominator.

    Returns:
        The result of the operation.

    Raises:
        ValueError: If any value is not an int or a string that follows variable_regex.pattern.
    """
    numerator = sympify_variable(numerator)
    denominator = sympify_variable(denominator)
    result = numerator/denominator
    return get_nonrational_variable(result)


def conv_out_length(length: Union[str, int], kernel: int, stride: int, padding: int, dilation: int) -> Union[str, int]:
    """Performs a symbolic calculation of the output length of a convolution.

    Args:
        length: Length of the input, either an int or a variable.
        kernel: Size of the kernel in the direction of length.
        stride: Stride size in the direction of the length.
        padding: Padding added at both sides in the direction of the length.
        dilation: Dilation size in the direction of the length.

    Returns:
        The result of the operation.
    """
    operation_sympy = sympify_variable('1+(length+2*padding-dilation*(kernel-1)-1)/stride')
    output_sympy = operation_sympy.subs({'length': sympify_variable(length),
                                         'kernel': sympify_variable(kernel),
                                         'stride': sympify_variable(stride),
                                         'padding': sympify_variable(padding),
                                         'dilation': sympify_variable(dilation)})
    frac_sympy = output_sympy.subs({s: 0 for s in output_sympy.free_symbols})
    output_sympy = output_sympy - frac_sympy + numbers.Integer(frac_sympy)
    return get_nonrational_variable(output_sympy)



"""Propagator classes for reshaping."""

from jsonargparse import Namespace
from jsonargparse import namespace_to_dict as n2d
from typing import List
from .base import BasePropagator, get_shape, create_shape
from ..sympy import prod, divide
from ..schemas import auto_tag, reshape_validator


def check_reshape_spec(reshape_spec):
    """Checks that reshape_spec is valid according to schema, indexes range is valid and there is at most one <<auto>> in each unflatten."""
    reshape_validator.validate(reshape_spec)
    idxs = []
    if reshape_spec != 'flatten':
        for val in reshape_spec:
            if isinstance(val, (int, str)):
                idxs.append(val)
            elif isinstance(val, list):
                idxs.extend([x for x in val])
            else:
                idx = next(iter(val.keys()))
                idxs.append(int(idx))
                if sum([x == auto_tag for x in val[idx]]) > 1:
                    raise ValueError(f'At most one {auto_tag} is allowed in unflatten definition ({val[idx]}).')
        if sorted(idxs) != list(range(len(idxs))):
            raise ValueError(f'Invalid indexes range ({sorted(idxs)}) in reshape_spec.')
    return idxs


def norm_reshape_spec(reshape_spec):
    """Converts elements of a reshape_spec from Namespace to dict."""
    if isinstance(reshape_spec, str):
        return reshape_spec
    return [n2d(x) if isinstance(x, Namespace) else x for x in reshape_spec]


class ReshapePropagator(BasePropagator):
    """Propagator for reshapping which could involve any of: permute, flatten and unflatten."""

    num_input_blocks = 1


    def initial_checks(self, from_blocks: List[Namespace], block: Namespace):
        """Method that does some initial checks before propagation.

        Calls the base class checks and makes sure that the reshape_spec
        attribute is valid and agrees with the input dimensions.

        Args:
            from_blocks: The input blocks.
            block: The block to propagate its shapes.

        Raises:
            ValueError: When block does not have a valid reshape_spec attribute that agrees with input dimensions.
        """
        super().initial_checks(from_blocks, block)
        if block.reshape_spec != 'flatten':
            reshape_spec = norm_reshape_spec(block.reshape_spec)
            try:
                idxs = check_reshape_spec(reshape_spec)
            except Exception as ex:
                raise ValueError(f'Invalid reshape_spec attribute in block[id={block._id}] :: {ex}') from ex
            shape_in = get_shape('out', from_blocks[0])
            if len(idxs) != len(shape_in):
                raise ValueError(f'Number of dimensions indexes in reshape_spec attribute of block[id={block._id}] does '
                                 f'not agree with the input dimensions coming from block[id={from_blocks[0]._id}].')


    def propagate(self, from_blocks: List[Namespace], block: Namespace):
        """Method that propagates shapes to a block.

        Args:
            from_blocks: The input blocks.
            block: The block to propagate its shapes.
        """
        shape_in = get_shape('out', from_blocks[0])
        shape_out = []
        if block.reshape_spec == 'flatten':
            reshape_spec = [[n for n in range(len(shape_in))]]
        else:
            reshape_spec = norm_reshape_spec(block.reshape_spec)
        for val in reshape_spec:
            if isinstance(val, int):
                shape_out.append(shape_in[val])
            elif isinstance(val, list):
                shape_out.append(prod([shape_in[x] for x in val]))
            elif isinstance(val, dict):
                idx = next(iter(val.keys()))
                in_dim = shape_in[int(idx)]
                dims = val[idx]
                if any(x == auto_tag for x in dims):
                    auto_idx = dims.index(auto_tag)
                    nonauto = prod([x for x in dims if x != auto_tag])
                    dims[auto_idx] = divide(in_dim, nonauto)
                shape_out.extend(dims)
        block._shape = create_shape(shape_in, shape_out)



import os

data_dir = os.path.dirname(os.path.realpath(__file__))

resnet_jsonnet = os.path.join(data_dir, 'resnet.jsonnet')
resnet_ext_vars = {'num_blocks': [2, 2, 2, 2]}
resnet_cfg = {'ext_vars': resnet_ext_vars}

resnet_multiscale_jsonnet = os.path.join(data_dir, 'resnet_multiscale.jsonnet')

squeezenet_jsonnet = os.path.join(data_dir, 'squeezenet.jsonnet')

text_image_jsonnet = os.path.join(data_dir, 'text_image_classification.jsonnet')
text_image_ext_vars = {'vocabulary_size': 5000}
text_image_cfg = {'ext_vars': text_image_ext_vars}

laia_jsonnet = os.path.join(data_dir, 'laia.jsonnet')
laia_ext_vars = {'num_symbols': 68}
laia_cfg = {'ext_vars': laia_ext_vars}
laia_shapes = [[16, 32, '<<variable:W/2>>'],
               [16, 16, '<<variable:W/4>>'],
               [32, 16, '<<variable:W/4>>'],
               [32, 8, '<<variable:W/8>>'],
               ['<<variable:W/8>>', 256],
               ['<<variable:W/8>>', 512],
               ['<<variable:W/8>>', 68]]

nested1_jsonnet = os.path.join(data_dir, 'nested1.jsonnet')
nested2_jsonnet = os.path.join(data_dir, 'nested/nested/nested2.jsonnet')
nested3_jsonnet = os.path.join(data_dir, 'nested/nested3.jsonnet')
nested1_ext_vars = {'hidden_size': 64, 'output_feats': 16}
nested2_ext_vars = {'input_size': 128, 'nested3_size': 64, 'output_feats': 16}
nested3_ext_vars = {'input_size': 64, 'output_feats': 16}
